cpdef tuple compare(prediction, reference, bint lenient=?, bint strict_strandedness=?, int fuzzymatch=?)
cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=?)
cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9])
