"""
This package contains the modules with the private methods employed by the Transcript class.
"""

from ..._transcripts.transcript_methods import finalizing, printing

__author__ = 'Luca Venturini'

__all__ = ["finalizing", "printing", "retrieval", "splitting"]
