__author__ = 'Luca Venturini'

from . import assigner_test
from . import intervaltree_test
from . import locus_test
from . import orf_test
from . import test_bed12
from . import test_blast_related
from . import test_clique_methods
from . import test_db_utils
from . import test_external_scores
from . import test_gene
from . import test_invalid_orfs
from . import test_metrics
from . import test_modifications
from . import test_scores
from . import test_serialise_junctions
from . import test_splitting
from . import test_system_calls
from . import test_transcript_checker
from . import test_transcript_methods
from . import *
from . import test_utils
from . import test_transcript_negative
from . import test_transcript_positive
from . import test_transcript_single
from . import utilities_test
