# coding: utf-8

"""
    This module defines the transcript-like objects.
"""

from .._transcripts.transcript_base import Namespace
from .transcript import Transcript, Metric
from .reference_gene import Gene
from .transcriptchecker import TranscriptChecker
from .transcriptcomputer import TranscriptComputer
