/* Generated by Cython 0.29.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "NPY_NO_DEPRECATED_API",
                "NPY_1_9_API_VERSION"
            ]
        ],
        "language": "c++",
        "name": "Mikado.scales.contrast",
        "sources": [
            "Mikado/scales/contrast.pyx"
        ]
    },
    "module_name": "Mikado.scales.contrast"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__Mikado__scales__contrast
#define __PYX_HAVE_API__Mikado__scales__contrast
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "Mikado/scales/contrast.pyx",
  "Mikado/utilities/intervaltree.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval;
struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode;
struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree;
struct __pyx_opt_args_6Mikado_9utilities_7overlap_overlap;

/* "utilities/overlap.pxd":1
 * cpdef long overlap(first, second, long flank=?, bint positive=?)             # <<<<<<<<<<<<<<
 * cdef long c_overlap(long start, long end, long ostart, long oend, long flank, bint positive)
 */
struct __pyx_opt_args_6Mikado_9utilities_7overlap_overlap {
  int __pyx_n;
  long flank;
  int positive;
};
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_left;
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_right;
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalTree_find;

/* "utilities/intervaltree.pxd":28
 *     cpdef intersect(IntervalNode self, int start, int stop)
 *     # cpdef right(self, Interval f, int n=1, int max_dist=25000)
 *     cpdef left(IntervalNode self, Interval f, int n=?, int max_dist=?, bint overlap=?)             # <<<<<<<<<<<<<<
 *     cpdef right(IntervalNode self, Interval f, int n=?, int max_dist=?, bint overlap=?)
 *     # cpdef left(self, Interval f, int n=1, int max_dist=25000)
 */
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_left {
  int __pyx_n;
  int n;
  int max_dist;
  int overlap;
};

/* "utilities/intervaltree.pxd":29
 *     # cpdef right(self, Interval f, int n=1, int max_dist=25000)
 *     cpdef left(IntervalNode self, Interval f, int n=?, int max_dist=?, bint overlap=?)
 *     cpdef right(IntervalNode self, Interval f, int n=?, int max_dist=?, bint overlap=?)             # <<<<<<<<<<<<<<
 *     # cpdef left(self, Interval f, int n=1, int max_dist=25000)
 *     # cpdef IntervalNode insert(IntervalNode self, int start, int end, object value=*)
 */
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_right {
  int __pyx_n;
  int n;
  int max_dist;
  int overlap;
};

/* "utilities/intervaltree.pxd":39
 *     cdef int num_intervals
 *     cpdef int size(IntervalTree self)
 *     cpdef find(IntervalTree self, int start, int end, bint strict=?,             # <<<<<<<<<<<<<<
 *                bint contained_check=?, int max_distance=?, int n=?, object value=?)
 *     # cpdef insert(IntervalTree self, int start, int end, object value=?)
 */
struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalTree_find {
  int __pyx_n;
  int strict;
  int contained_check;
  int max_distance;
  int n;
  PyObject *value;
};
struct __pyx_opt_args_6Mikado_6scales_8contrast_compare;
struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode;

/* "Mikado/scales/contrast.pxd":1
 * cpdef tuple compare(prediction, reference, bint lenient=?, bint strict_strandedness=?, int fuzzymatch=?)             # <<<<<<<<<<<<<<
 * cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=?)
 * cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9])
 */
struct __pyx_opt_args_6Mikado_6scales_8contrast_compare {
  int __pyx_n;
  int lenient;
  int strict_strandedness;
  int fuzzymatch;
};

/* "Mikado/scales/contrast.pxd":2
 * cpdef tuple compare(prediction, reference, bint lenient=?, bint strict_strandedness=?, int fuzzymatch=?)
 * cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=?)             # <<<<<<<<<<<<<<
 * cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9])
 */
struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode {
  int __pyx_n;
  int fuzzymatch;
};

/* "utilities/intervaltree.pxd":3
 * #distutils: language = c++
 * 
 * cdef class Interval:             # <<<<<<<<<<<<<<
 *     cdef public int start, end
 *     cdef public object data
 */
struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval {
  PyObject_HEAD
  struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_Interval *__pyx_vtab;
  int start;
  int end;
  PyObject *data;
  PyObject *value;
  PyObject *chrom;
  PyObject *strand;
};


/* "utilities/intervaltree.pxd":11
 * 
 * 
 * cdef class IntervalNode:             # <<<<<<<<<<<<<<
 *     cdef float priority
 *     cdef public Interval interval
 */
struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalNode *__pyx_vtab;
  float priority;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *interval;
  int start;
  int end;
  int minstop;
  int maxstop;
  int minstart;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *cleft;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *cright;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *croot;
};


/* "utilities/intervaltree.pxd":35
 * 
 * 
 * cdef class IntervalTree:             # <<<<<<<<<<<<<<
 *     cdef IntervalNode root
 *     cdef int num_intervals
 */
struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree {
  PyObject_HEAD
  struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree *__pyx_vtab;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *root;
  int num_intervals;
};



/* "utilities/intervaltree.pxd":3
 * #distutils: language = c++
 * 
 * cdef class Interval:             # <<<<<<<<<<<<<<
 *     cdef public int start, end
 *     cdef public object data
 */

struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_Interval {
  PyObject *(*_as_tuple)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_Interval *__pyx_vtabptr_6Mikado_9utilities_12intervaltree_Interval;


/* "utilities/intervaltree.pxd":11
 * 
 * 
 * cdef class IntervalNode:             # <<<<<<<<<<<<<<
 *     cdef float priority
 *     cdef public Interval interval
 */

struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalNode {
  PyObject *(*_as_tuple)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int __pyx_skip_dispatch);
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *(*rotate_right)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *);
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *(*rotate_left)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *);
  void (*set_stops)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *);
  void (*_intersect)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int, int, PyObject *);
  void (*_seek_left)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int, PyObject *, int, int);
  void (*_seek_right)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int, PyObject *, int, int);
  void (*_traverse)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, PyObject *);
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *(*_insert)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *);
  PyObject *(*intersect)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int, int, int __pyx_skip_dispatch);
  PyObject *(*left)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *, int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_left *__pyx_optional_args);
  PyObject *(*right)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *, int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalNode_right *__pyx_optional_args);
  int (*fuzzy_equal)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalNode *__pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalNode;


/* "utilities/intervaltree.pxd":35
 * 
 * 
 * cdef class IntervalTree:             # <<<<<<<<<<<<<<
 *     cdef IntervalNode root
 *     cdef int num_intervals
 */

struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree {
  int (*size)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *, int __pyx_skip_dispatch);
  PyObject *(*find)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *, int, int, int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalTree_find *__pyx_optional_args);
  PyObject *(*insert)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *, struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval *, int __pyx_skip_dispatch);
  void (*set_stops)(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *);
};
static struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree *__pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalTree;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* Profile.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #define __Pyx_TraceDeclarations\
      static PyCodeObject *__pyx_frame_code = NULL;\
      CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
      int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
      if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
#if PY_VERSION_HEX >= 0x030a00b1
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_SetTracing(tstate, enable)\
      (tstate)->cframe->use_tracing = (enable)
#else
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_SetTracing(tstate, enable)\
      (tstate)->use_tracing = (enable)
#endif
  #ifdef WITH_THREAD
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 1, 1)) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #else
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)\
  {   PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall(&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #endif
  #define __Pyx_TraceException()\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 1)) {\
          tstate->tracing++;\
          __Pyx_SetTracing(tstate, 0);\
          PyObject *exc_info = __Pyx_GetExceptionTuple(tstate);\
          if (exc_info) {\
              if (CYTHON_TRACE && tstate->c_tracefunc)\
                  tstate->c_tracefunc(\
                      tstate->c_traceobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              tstate->c_profilefunc(\
                  tstate->c_profileobj, __pyx_frame, PyTrace_EXCEPTION, exc_info);\
              Py_DECREF(exc_info);\
          }\
          __Pyx_SetTracing(tstate, 1);\
          tstate->tracing--;\
      }\
  }
  static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result) {
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      tstate->tracing++;
      __Pyx_SetTracing(tstate, 0);
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      CYTHON_FRAME_DEL(frame);
      __Pyx_SetTracing(tstate, 1);
      tstate->tracing--;
      __Pyx_ErrRestoreInState(tstate, type, value, traceback);
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
          }\
      }\
  }
  #else
  #define __Pyx_TraceReturn(result, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0)) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result);\
      }\
  }
  #endif
  static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno);
#else
  #define __Pyx_TraceDeclarations
  #define __Pyx_TraceFrameInit(codeobj)
  #define __Pyx_TraceCall(funcname, srcfile, firstlineno, nogil, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceException()
  #define __Pyx_TraceReturn(result, nogil)
#endif
#if CYTHON_TRACE
  static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int lineno) {
      int ret;
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_PyFrame_SetLineNumber(frame, lineno);
      tstate->tracing++;
      __Pyx_SetTracing(tstate, 0);
      ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
      __Pyx_SetTracing(tstate, 1);
      tstate->tracing--;
      if (likely(!ret)) {
          __Pyx_ErrRestoreInState(tstate, type, value, traceback);
      } else {
          Py_XDECREF(type);
          Py_XDECREF(value);
          Py_XDECREF(traceback);
      }
      return ret;
  }
  #ifdef WITH_THREAD
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              int ret = 0;\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              }\
              PyGILState_Release(state);\
              if (unlikely(ret)) goto_error;\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
              int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
              if (unlikely(ret)) goto_error;\
          }\
      }\
  }
  #else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
          int ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, lineno);\
          if (unlikely(ret)) goto_error;\
      }\
  }
  #endif
#else
  #define __Pyx_TraceLine(lineno, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* py_set_remove.proto */
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'Mikado.utilities.overlap' */
static long (*__pyx_f_6Mikado_9utilities_7overlap_c_overlap)(long, long, long, long, long, int); /*proto*/

/* Module declarations from 'Mikado.utilities.intervaltree' */
static PyTypeObject *__pyx_ptype_6Mikado_9utilities_12intervaltree_Interval = 0;
static PyTypeObject *__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalNode = 0;
static PyTypeObject *__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree = 0;

/* Module declarations from 'Mikado.utilities.f1' */
static double (*__pyx_f_6Mikado_9utilities_2f1_calc_f1)(double, double, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'Mikado.scales.contrast' */
static PyObject *__pyx_f_6Mikado_6scales_8contrast_compare(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_6scales_8contrast_compare *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_6Mikado_6scales_8contrast___assign_multiexonic_ccode(PyObject *, PyObject *, long, double *, struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_6Mikado_6scales_8contrast___assign_monoexonic_ccode(PyObject *, PyObject *, long, double *); /*proto*/
static CYTHON_INLINE long __pyx_f_6Mikado_6scales_8contrast_long_max(long, long); /*proto*/
static CYTHON_INLINE double __pyx_f_6Mikado_6scales_8contrast_double_min(double, double); /*proto*/
#define __Pyx_MODULE_NAME "Mikado.scales.contrast"
extern int __pyx_module_is_main_Mikado__scales__contrast;
int __pyx_module_is_main_Mikado__scales__contrast = 0;

/* Implementation of 'Mikado.scales.contrast' */
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_[] = "";
static const char __pyx_k_C[] = "C";
static const char __pyx_k_G[] = "G";
static const char __pyx_k_I[] = "I";
static const char __pyx_k_J[] = "J";
static const char __pyx_k_P[] = "P";
static const char __pyx_k_X[] = "X";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_u[] = "u";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_NA[] = "NA";
static const char __pyx_k__2[] = "_";
static const char __pyx_k__3[] = "=";
static const char __pyx_k__4[] = "{}:{}..{}";
static const char __pyx_k__5[] = ",";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_rI[] = "rI";
static const char __pyx_k_ri[] = "ri";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_exon[] = "exon";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_chrom[] = "chrom";
static const char __pyx_k_exons[] = "exons";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_author[] = "__author__";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_intron[] = "intron";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_strand[] = "strand";
static const char __pyx_k_introns[] = "introns";
static const char __pyx_k_lenient[] = "lenient";
static const char __pyx_k_splices[] = "splices";
static const char __pyx_k_exon_num[] = "exon_num";
static const char __pyx_k_finalize[] = "finalize";
static const char __pyx_k_reference[] = "reference";
static const char __pyx_k_Transcript[] = "Transcript";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_difference[] = "difference";
static const char __pyx_k_fuzzymatch[] = "fuzzymatch";
static const char __pyx_k_prediction[] = "prediction";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_cdna_length[] = "cdna_length";
static const char __pyx_k_segmenttree[] = "segmenttree";
static const char __pyx_k_transcripts[] = "transcripts";
static const char __pyx_k_ResultStorer[] = "ResultStorer";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_resultstorer[] = "resultstorer";
static const char __pyx_k_Luca_Venturini[] = "Luca Venturini";
static const char __pyx_k_Ccode_is_null_0[] = "Ccode is null;\n{0}";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strict_strandedness[] = "strict_strandedness";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_n_u_C;
static PyObject *__pyx_kp_u_Ccode_is_null_0;
static PyObject *__pyx_n_u_G;
static PyObject *__pyx_n_u_I;
static PyObject *__pyx_n_u_J;
static PyObject *__pyx_kp_u_Luca_Venturini;
static PyObject *__pyx_n_u_NA;
static PyObject *__pyx_n_u_P;
static PyObject *__pyx_n_s_ResultStorer;
static PyObject *__pyx_n_s_Transcript;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_u_X;
static PyObject *__pyx_n_u__2;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_n_s_author;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_cdna_length;
static PyObject *__pyx_n_s_chrom;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_difference;
static PyObject *__pyx_n_u_e;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_u_exon;
static PyObject *__pyx_n_s_exon_num;
static PyObject *__pyx_n_s_exons;
static PyObject *__pyx_n_s_finalize;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fuzzymatch;
static PyObject *__pyx_n_u_g;
static PyObject *__pyx_n_u_h;
static PyObject *__pyx_n_u_i;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_intersection;
static PyObject *__pyx_n_u_intron;
static PyObject *__pyx_n_s_introns;
static PyObject *__pyx_n_u_j;
static PyObject *__pyx_n_s_lenient;
static PyObject *__pyx_n_u_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_u_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_u_o;
static PyObject *__pyx_n_u_p;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_prediction;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_u_rI;
static PyObject *__pyx_n_s_reference;
static PyObject *__pyx_n_s_resultstorer;
static PyObject *__pyx_n_u_ri;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_s_segmenttree;
static PyObject *__pyx_n_s_splices;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_strand;
static PyObject *__pyx_n_s_strict_strandedness;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_transcripts;
static PyObject *__pyx_n_u_u;
static PyObject *__pyx_n_u_x;
static PyObject *__pyx_pf_6Mikado_6scales_8contrast_compare(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_prediction, PyObject *__pyx_v_reference, int __pyx_v_lenient, int __pyx_v_strict_strandedness, int __pyx_v_fuzzymatch); /* proto */
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_copy = {0, &__pyx_n_s_copy, 0, 0, 0};
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
/* Late includes */

/* "Mikado/scales/contrast.pyx":11
 * __author__ = 'Luca Venturini'
 * 
 * cdef inline long long_max(long a, long b): return a if a >= b else b             # <<<<<<<<<<<<<<
 * cdef inline double double_min(double a, double b): return a if a <= b else b
 * 
 */

static CYTHON_INLINE long __pyx_f_6Mikado_6scales_8contrast_long_max(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("long_max", 0);
  if (((__pyx_v_a >= __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Mikado/scales/contrast.pyx":12
 * 
 * cdef inline long long_max(long a, long b): return a if a >= b else b
 * cdef inline double double_min(double a, double b): return a if a <= b else b             # <<<<<<<<<<<<<<
 * 
 * @cython.profile(True)
 */

static CYTHON_INLINE double __pyx_f_6Mikado_6scales_8contrast_double_min(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("double_min", 0);
  if (((__pyx_v_a <= __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Mikado/scales/contrast.pyx":16
 * @cython.profile(True)
 * @cython.boundscheck(False)
 * cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9]):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

static PyObject *__pyx_f_6Mikado_6scales_8contrast___assign_monoexonic_ccode(PyObject *__pyx_v_prediction, PyObject *__pyx_v_reference, long __pyx_v_nucl_overlap, double *__pyx_v_stats) {
  PyObject *__pyx_v_ccode = 0;
  double __pyx_v_nucl_recall;
  double __pyx_v_nucl_precision;
  double __pyx_v_nucl_f1;
  CYTHON_UNUSED double __pyx_v_exon_recall;
  CYTHON_UNUSED double __pyx_v_exon_precision;
  CYTHON_UNUSED double __pyx_v_exon_f1;
  CYTHON_UNUSED double __pyx_v_junction_recall;
  CYTHON_UNUSED double __pyx_v_junction_precision;
  CYTHON_UNUSED double __pyx_v_junction_f1;
  short __pyx_v_p_exon_num;
  short __pyx_v_r_exon_num;
  long __pyx_v_p_start;
  long __pyx_v_p_end;
  long __pyx_v_r_start;
  long __pyx_v_r_end;
  long __pyx_v_over;
  long __pyx_v_over_left;
  long __pyx_v_over_right;
  CYTHON_UNUSED long __pyx_v_p_cdna_length;
  long __pyx_v_i_length;
  long __pyx_v_i_start;
  long __pyx_v_i_end;
  long __pyx_v_introns[2];
  long __pyx_v_overlaps[3];
  unsigned short __pyx_v_index;
  PyObject *__pyx_v_r_strand = 0;
  PyObject *__pyx_v_p_strand = 0;
  PyObject *__pyx_v_intron = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_reference_exon = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  short __pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__assign_monoexonic_ccode", 0);
  __Pyx_TraceCall("__assign_monoexonic_ccode", __pyx_f[0], 16, 0, __PYX_ERR(0, 16, __pyx_L1_error));

  /* "Mikado/scales/contrast.pyx":35
 *         # double stats[9]
 * 
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end             # <<<<<<<<<<<<<<
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 *     if reference.strand is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_short(__pyx_t_1); if (unlikely((__pyx_t_2 == (short)-1) && PyErr_Occurred())) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_exon_num = __pyx_t_2;
  __pyx_v_p_start = __pyx_t_3;
  __pyx_v_p_end = __pyx_t_4;

  /* "Mikado/scales/contrast.pyx":36
 * 
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end             # <<<<<<<<<<<<<<
 *     if reference.strand is None:
 *         r_strand = ""
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_short(__pyx_t_1); if (unlikely((__pyx_t_2 == (short)-1) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_r_exon_num = __pyx_t_2;
  __pyx_v_r_start = __pyx_t_4;
  __pyx_v_r_end = __pyx_t_3;

  /* "Mikado/scales/contrast.pyx":37
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 *     if reference.strand is None:             # <<<<<<<<<<<<<<
 *         r_strand = ""
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "Mikado/scales/contrast.pyx":38
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 *     if reference.strand is None:
 *         r_strand = ""             # <<<<<<<<<<<<<<
 *     else:
 *         r_strand = reference.strand
 */
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_v_r_strand = __pyx_kp_u_;

    /* "Mikado/scales/contrast.pyx":37
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 *     if reference.strand is None:             # <<<<<<<<<<<<<<
 *         r_strand = ""
 *     else:
 */
    goto __pyx_L3;
  }

  /* "Mikado/scales/contrast.pyx":40
 *         r_strand = ""
 *     else:
 *         r_strand = reference.strand             # <<<<<<<<<<<<<<
 * 
 *     if prediction.strand is None:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 40, __pyx_L1_error)
    __pyx_v_r_strand = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Mikado/scales/contrast.pyx":42
 *         r_strand = reference.strand
 * 
 *     if prediction.strand is None:             # <<<<<<<<<<<<<<
 *         p_strand = ""
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "Mikado/scales/contrast.pyx":43
 * 
 *     if prediction.strand is None:
 *         p_strand = ""             # <<<<<<<<<<<<<<
 *     else:
 *         p_strand = prediction.strand
 */
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_v_p_strand = __pyx_kp_u_;

    /* "Mikado/scales/contrast.pyx":42
 *         r_strand = reference.strand
 * 
 *     if prediction.strand is None:             # <<<<<<<<<<<<<<
 *         p_strand = ""
 *     else:
 */
    goto __pyx_L4;
  }

  /* "Mikado/scales/contrast.pyx":45
 *         p_strand = ""
 *     else:
 *         p_strand = prediction.strand             # <<<<<<<<<<<<<<
 * 
 *     ccode = ""
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 45, __pyx_L1_error)
    __pyx_v_p_strand = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "Mikado/scales/contrast.pyx":47
 *         p_strand = prediction.strand
 * 
 *     ccode = ""             # <<<<<<<<<<<<<<
 *     nucl_recall = stats[0]
 *     nucl_precision = stats[1]
 */
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_v_ccode = __pyx_kp_u_;

  /* "Mikado/scales/contrast.pyx":48
 * 
 *     ccode = ""
 *     nucl_recall = stats[0]             # <<<<<<<<<<<<<<
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]
 */
  __pyx_v_nucl_recall = (__pyx_v_stats[0]);

  /* "Mikado/scales/contrast.pyx":49
 *     ccode = ""
 *     nucl_recall = stats[0]
 *     nucl_precision = stats[1]             # <<<<<<<<<<<<<<
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]
 */
  __pyx_v_nucl_precision = (__pyx_v_stats[1]);

  /* "Mikado/scales/contrast.pyx":50
 *     nucl_recall = stats[0]
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]             # <<<<<<<<<<<<<<
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]
 */
  __pyx_v_nucl_f1 = (__pyx_v_stats[2]);

  /* "Mikado/scales/contrast.pyx":51
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]             # <<<<<<<<<<<<<<
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]
 */
  __pyx_v_exon_recall = (__pyx_v_stats[3]);

  /* "Mikado/scales/contrast.pyx":52
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]             # <<<<<<<<<<<<<<
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]
 */
  __pyx_v_exon_precision = (__pyx_v_stats[4]);

  /* "Mikado/scales/contrast.pyx":53
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]             # <<<<<<<<<<<<<<
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]
 */
  __pyx_v_exon_f1 = (__pyx_v_stats[5]);

  /* "Mikado/scales/contrast.pyx":54
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]             # <<<<<<<<<<<<<<
 *     junction_precision = stats[7]
 *     junction_f1 = stats[8]
 */
  __pyx_v_junction_recall = (__pyx_v_stats[6]);

  /* "Mikado/scales/contrast.pyx":55
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]             # <<<<<<<<<<<<<<
 *     junction_f1 = stats[8]
 * 
 */
  __pyx_v_junction_precision = (__pyx_v_stats[7]);

  /* "Mikado/scales/contrast.pyx":56
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]
 *     junction_f1 = stats[8]             # <<<<<<<<<<<<<<
 * 
 *     if p_exon_num == 1 and r_exon_num > 1:
 */
  __pyx_v_junction_f1 = (__pyx_v_stats[8]);

  /* "Mikado/scales/contrast.pyx":58
 *     junction_f1 = stats[8]
 * 
 *     if p_exon_num == 1 and r_exon_num > 1:             # <<<<<<<<<<<<<<
 *         if nucl_precision < 1 and nucl_overlap > 0:
 *             index = -1
 */
  __pyx_t_6 = ((__pyx_v_p_exon_num == 1) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_r_exon_num > 1) != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Mikado/scales/contrast.pyx":59
 * 
 *     if p_exon_num == 1 and r_exon_num > 1:
 *         if nucl_precision < 1 and nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             index = -1
 *             for intron in sorted(reference.introns):
 */
    __pyx_t_6 = ((__pyx_v_nucl_precision < 1.0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_6 = ((__pyx_v_nucl_overlap > 0) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":60
 *     if p_exon_num == 1 and r_exon_num > 1:
 *         if nucl_precision < 1 and nucl_overlap > 0:
 *             index = -1             # <<<<<<<<<<<<<<
 *             for intron in sorted(reference.introns):
 *                 i_start, i_end = intron[0], intron[1]
 */
      __pyx_v_index = -1;

      /* "Mikado/scales/contrast.pyx":61
 *         if nucl_precision < 1 and nucl_overlap > 0:
 *             index = -1
 *             for intron in sorted(reference.introns):             # <<<<<<<<<<<<<<
 *                 i_start, i_end = intron[0], intron[1]
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_introns); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PySequence_List(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 61, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_1 = ((PyObject*)__pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_9 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 61, __pyx_L1_error)
      if (unlikely(__pyx_t_1 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 61, __pyx_L1_error)
      }
      __pyx_t_8 = __pyx_t_1; __Pyx_INCREF(__pyx_t_8); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 61, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_intron, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Mikado/scales/contrast.pyx":62
 *             index = -1
 *             for intron in sorted(reference.introns):
 *                 i_start, i_end = intron[0], intron[1]             # <<<<<<<<<<<<<<
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 *                 if over > 0:
 */
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_intron, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_3 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_intron, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_i_start = __pyx_t_3;
        __pyx_v_i_end = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":63
 *             for intron in sorted(reference.introns):
 *                 i_start, i_end = intron[0], intron[1]
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)             # <<<<<<<<<<<<<<
 *                 if over > 0:
 *                     index += 1
 */
        __pyx_v_over = __pyx_f_6Mikado_9utilities_7overlap_c_overlap(__pyx_v_i_start, __pyx_v_i_end, __pyx_v_p_start, __pyx_v_p_end, 0, 1);

        /* "Mikado/scales/contrast.pyx":64
 *                 i_start, i_end = intron[0], intron[1]
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 *                 if over > 0:             # <<<<<<<<<<<<<<
 *                     index += 1
 *                     if index > 1:
 */
        __pyx_t_5 = ((__pyx_v_over > 0) != 0);
        if (__pyx_t_5) {

          /* "Mikado/scales/contrast.pyx":65
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 *                 if over > 0:
 *                     index += 1             # <<<<<<<<<<<<<<
 *                     if index > 1:
 *                         break
 */
          __pyx_v_index = (__pyx_v_index + 1);

          /* "Mikado/scales/contrast.pyx":66
 *                 if over > 0:
 *                     index += 1
 *                     if index > 1:             # <<<<<<<<<<<<<<
 *                         break
 *                     i_length = i_end - i_start + 1
 */
          __pyx_t_5 = ((__pyx_v_index > 1) != 0);
          if (__pyx_t_5) {

            /* "Mikado/scales/contrast.pyx":67
 *                     index += 1
 *                     if index > 1:
 *                         break             # <<<<<<<<<<<<<<
 *                     i_length = i_end - i_start + 1
 *                     overlaps[index] = over
 */
            goto __pyx_L12_break;

            /* "Mikado/scales/contrast.pyx":66
 *                 if over > 0:
 *                     index += 1
 *                     if index > 1:             # <<<<<<<<<<<<<<
 *                         break
 *                     i_length = i_end - i_start + 1
 */
          }

          /* "Mikado/scales/contrast.pyx":68
 *                     if index > 1:
 *                         break
 *                     i_length = i_end - i_start + 1             # <<<<<<<<<<<<<<
 *                     overlaps[index] = over
 *                     introns[index] = i_length
 */
          __pyx_v_i_length = ((__pyx_v_i_end - __pyx_v_i_start) + 1);

          /* "Mikado/scales/contrast.pyx":69
 *                         break
 *                     i_length = i_end - i_start + 1
 *                     overlaps[index] = over             # <<<<<<<<<<<<<<
 *                     introns[index] = i_length
 * 
 */
          (__pyx_v_overlaps[__pyx_v_index]) = __pyx_v_over;

          /* "Mikado/scales/contrast.pyx":70
 *                     i_length = i_end - i_start + 1
 *                     overlaps[index] = over
 *                     introns[index] = i_length             # <<<<<<<<<<<<<<
 * 
 *             if index == -1:
 */
          (__pyx_v_introns[__pyx_v_index]) = __pyx_v_i_length;

          /* "Mikado/scales/contrast.pyx":64
 *                 i_start, i_end = intron[0], intron[1]
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 *                 if over > 0:             # <<<<<<<<<<<<<<
 *                     index += 1
 *                     if index > 1:
 */
        }

        /* "Mikado/scales/contrast.pyx":61
 *         if nucl_precision < 1 and nucl_overlap > 0:
 *             index = -1
 *             for intron in sorted(reference.introns):             # <<<<<<<<<<<<<<
 *                 i_start, i_end = intron[0], intron[1]
 *                 over = c_overlap(i_start, i_end, p_start, p_end, flank=0, positive=True)
 */
      }
      __pyx_L12_break:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Mikado/scales/contrast.pyx":72
 *                     introns[index] = i_length
 * 
 *             if index == -1:             # <<<<<<<<<<<<<<
 *                 # Completely contained inside
 *                 ccode = "g"
 */
      __pyx_t_5 = ((__pyx_v_index == -1L) != 0);
      if (__pyx_t_5) {

        /* "Mikado/scales/contrast.pyx":74
 *             if index == -1:
 *                 # Completely contained inside
 *                 ccode = "g"             # <<<<<<<<<<<<<<
 *             elif index == 0:
 *                 over = overlaps[0]
 */
        __Pyx_INCREF(__pyx_n_u_g);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_g);

        /* "Mikado/scales/contrast.pyx":72
 *                     introns[index] = i_length
 * 
 *             if index == -1:             # <<<<<<<<<<<<<<
 *                 # Completely contained inside
 *                 ccode = "g"
 */
        goto __pyx_L15;
      }

      /* "Mikado/scales/contrast.pyx":75
 *                 # Completely contained inside
 *                 ccode = "g"
 *             elif index == 0:             # <<<<<<<<<<<<<<
 *                 over = overlaps[0]
 *                 i_length = introns[0]
 */
      __pyx_t_5 = ((__pyx_v_index == 0) != 0);
      if (__pyx_t_5) {

        /* "Mikado/scales/contrast.pyx":76
 *                 ccode = "g"
 *             elif index == 0:
 *                 over = overlaps[0]             # <<<<<<<<<<<<<<
 *                 i_length = introns[0]
 *                 # over, i_length = overlaps[0][0], overlaps[0][1]
 */
        __pyx_v_over = (__pyx_v_overlaps[0]);

        /* "Mikado/scales/contrast.pyx":77
 *             elif index == 0:
 *                 over = overlaps[0]
 *                 i_length = introns[0]             # <<<<<<<<<<<<<<
 *                 # over, i_length = overlaps[0][0], overlaps[0][1]
 *                 p_cdna_length = prediction.cdna_length
 */
        __pyx_v_i_length = (__pyx_v_introns[0]);

        /* "Mikado/scales/contrast.pyx":79
 *                 i_length = introns[0]
 *                 # over, i_length = overlaps[0][0], overlaps[0][1]
 *                 p_cdna_length = prediction.cdna_length             # <<<<<<<<<<<<<<
 *                 if (10 < over < i_length) and (r_start < p_start < p_end < r_end):
 *                     ccode = "e"
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_cdna_length); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_8); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 79, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_v_p_cdna_length = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":80
 *                 # over, i_length = overlaps[0][0], overlaps[0][1]
 *                 p_cdna_length = prediction.cdna_length
 *                 if (10 < over < i_length) and (r_start < p_start < p_end < r_end):             # <<<<<<<<<<<<<<
 *                     ccode = "e"
 *                 else:
 */
        __pyx_t_6 = (10 < __pyx_v_over);
        if (__pyx_t_6) {
          __pyx_t_6 = (__pyx_v_over < __pyx_v_i_length);
        }
        __pyx_t_11 = (__pyx_t_6 != 0);
        if (__pyx_t_11) {
        } else {
          __pyx_t_5 = __pyx_t_11;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_11 = (__pyx_v_r_start < __pyx_v_p_start);
        if (__pyx_t_11) {
          __pyx_t_11 = (__pyx_v_p_start < __pyx_v_p_end);
          if (__pyx_t_11) {
            __pyx_t_11 = (__pyx_v_p_end < __pyx_v_r_end);
          }
        }
        __pyx_t_6 = (__pyx_t_11 != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L17_bool_binop_done:;
        if (__pyx_t_5) {

          /* "Mikado/scales/contrast.pyx":81
 *                 p_cdna_length = prediction.cdna_length
 *                 if (10 < over < i_length) and (r_start < p_start < p_end < r_end):
 *                     ccode = "e"             # <<<<<<<<<<<<<<
 *                 else:
 *                     ccode = "g"
 */
          __Pyx_INCREF(__pyx_n_u_e);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_e);

          /* "Mikado/scales/contrast.pyx":80
 *                 # over, i_length = overlaps[0][0], overlaps[0][1]
 *                 p_cdna_length = prediction.cdna_length
 *                 if (10 < over < i_length) and (r_start < p_start < p_end < r_end):             # <<<<<<<<<<<<<<
 *                     ccode = "e"
 *                 else:
 */
          goto __pyx_L16;
        }

        /* "Mikado/scales/contrast.pyx":83
 *                     ccode = "e"
 *                 else:
 *                     ccode = "g"             # <<<<<<<<<<<<<<
 *             elif index > 1:
 *                 ccode = "g"
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_n_u_g);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_g);
        }
        __pyx_L16:;

        /* "Mikado/scales/contrast.pyx":75
 *                 # Completely contained inside
 *                 ccode = "g"
 *             elif index == 0:             # <<<<<<<<<<<<<<
 *                 over = overlaps[0]
 *                 i_length = introns[0]
 */
        goto __pyx_L15;
      }

      /* "Mikado/scales/contrast.pyx":84
 *                 else:
 *                     ccode = "g"
 *             elif index > 1:             # <<<<<<<<<<<<<<
 *                 ccode = "g"
 *             elif index == 1:
 */
      __pyx_t_5 = ((__pyx_v_index > 1) != 0);
      if (__pyx_t_5) {

        /* "Mikado/scales/contrast.pyx":85
 *                     ccode = "g"
 *             elif index > 1:
 *                 ccode = "g"             # <<<<<<<<<<<<<<
 *             elif index == 1:
 *                 over_left, over_right = overlaps[0], overlaps[1]
 */
        __Pyx_INCREF(__pyx_n_u_g);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_g);

        /* "Mikado/scales/contrast.pyx":84
 *                 else:
 *                     ccode = "g"
 *             elif index > 1:             # <<<<<<<<<<<<<<
 *                 ccode = "g"
 *             elif index == 1:
 */
        goto __pyx_L15;
      }

      /* "Mikado/scales/contrast.pyx":86
 *             elif index > 1:
 *                 ccode = "g"
 *             elif index == 1:             # <<<<<<<<<<<<<<
 *                 over_left, over_right = overlaps[0], overlaps[1]
 *                 if over_left < 10 and over_right < 10:
 */
      __pyx_t_5 = ((__pyx_v_index == 1) != 0);
      if (__pyx_t_5) {

        /* "Mikado/scales/contrast.pyx":87
 *                 ccode = "g"
 *             elif index == 1:
 *                 over_left, over_right = overlaps[0], overlaps[1]             # <<<<<<<<<<<<<<
 *                 if over_left < 10 and over_right < 10:
 *                     ccode = "g"
 */
        __pyx_t_4 = (__pyx_v_overlaps[0]);
        __pyx_t_3 = (__pyx_v_overlaps[1]);
        __pyx_v_over_left = __pyx_t_4;
        __pyx_v_over_right = __pyx_t_3;

        /* "Mikado/scales/contrast.pyx":88
 *             elif index == 1:
 *                 over_left, over_right = overlaps[0], overlaps[1]
 *                 if over_left < 10 and over_right < 10:             # <<<<<<<<<<<<<<
 *                     ccode = "g"
 *                 else:
 */
        __pyx_t_6 = ((__pyx_v_over_left < 10) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_5 = __pyx_t_6;
          goto __pyx_L20_bool_binop_done;
        }
        __pyx_t_6 = ((__pyx_v_over_right < 10) != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L20_bool_binop_done:;
        if (__pyx_t_5) {

          /* "Mikado/scales/contrast.pyx":89
 *                 over_left, over_right = overlaps[0], overlaps[1]
 *                 if over_left < 10 and over_right < 10:
 *                     ccode = "g"             # <<<<<<<<<<<<<<
 *                 else:
 *                     ccode = "e"
 */
          __Pyx_INCREF(__pyx_n_u_g);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_g);

          /* "Mikado/scales/contrast.pyx":88
 *             elif index == 1:
 *                 over_left, over_right = overlaps[0], overlaps[1]
 *                 if over_left < 10 and over_right < 10:             # <<<<<<<<<<<<<<
 *                     ccode = "g"
 *                 else:
 */
          goto __pyx_L19;
        }

        /* "Mikado/scales/contrast.pyx":91
 *                     ccode = "g"
 *                 else:
 *                     ccode = "e"             # <<<<<<<<<<<<<<
 *         elif nucl_precision == 1:
 *             ccode = "c"
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_n_u_e);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_e);
        }
        __pyx_L19:;

        /* "Mikado/scales/contrast.pyx":86
 *             elif index > 1:
 *                 ccode = "g"
 *             elif index == 1:             # <<<<<<<<<<<<<<
 *                 over_left, over_right = overlaps[0], overlaps[1]
 *                 if over_left < 10 and over_right < 10:
 */
      }
      __pyx_L15:;

      /* "Mikado/scales/contrast.pyx":59
 * 
 *     if p_exon_num == 1 and r_exon_num > 1:
 *         if nucl_precision < 1 and nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             index = -1
 *             for intron in sorted(reference.introns):
 */
      goto __pyx_L8;
    }

    /* "Mikado/scales/contrast.pyx":92
 *                 else:
 *                     ccode = "e"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"
 *         elif nucl_overlap > 0:
 */
    __pyx_t_5 = ((__pyx_v_nucl_precision == 1.0) != 0);
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":93
 *                     ccode = "e"
 *         elif nucl_precision == 1:
 *             ccode = "c"             # <<<<<<<<<<<<<<
 *         elif nucl_overlap > 0:
 *             ccode = "g"
 */
      __Pyx_INCREF(__pyx_n_u_c);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_c);

      /* "Mikado/scales/contrast.pyx":92
 *                 else:
 *                     ccode = "e"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"
 *         elif nucl_overlap > 0:
 */
      goto __pyx_L8;
    }

    /* "Mikado/scales/contrast.pyx":94
 *         elif nucl_precision == 1:
 *             ccode = "c"
 *         elif nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             ccode = "g"
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 */
    __pyx_t_5 = ((__pyx_v_nucl_overlap > 0) != 0);
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":95
 *             ccode = "c"
 *         elif nucl_overlap > 0:
 *             ccode = "g"             # <<<<<<<<<<<<<<
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 *             ccode = "i"  # Monoexonic fragment inside an intron
 */
      __Pyx_INCREF(__pyx_n_u_g);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_g);

      /* "Mikado/scales/contrast.pyx":94
 *         elif nucl_precision == 1:
 *             ccode = "c"
 *         elif nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             ccode = "g"
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 */
      goto __pyx_L8;
    }

    /* "Mikado/scales/contrast.pyx":96
 *         elif nucl_overlap > 0:
 *             ccode = "g"
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):             # <<<<<<<<<<<<<<
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:
 */
    __pyx_t_6 = ((__pyx_v_nucl_recall == 0.0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_r_start < __pyx_v_p_start);
    if (__pyx_t_6) {
      __pyx_t_6 = (__pyx_v_p_start < __pyx_v_r_end);
    }
    __pyx_t_11 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_11;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":97
 *             ccode = "g"
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 *             ccode = "i"  # Monoexonic fragment inside an intron             # <<<<<<<<<<<<<<
 *     elif p_exon_num > 1 and r_exon_num == 1:
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 */
      __Pyx_INCREF(__pyx_n_u_i);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_i);

      /* "Mikado/scales/contrast.pyx":96
 *         elif nucl_overlap > 0:
 *             ccode = "g"
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):             # <<<<<<<<<<<<<<
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:
 */
    }
    __pyx_L8:;

    /* "Mikado/scales/contrast.pyx":58
 *     junction_f1 = stats[8]
 * 
 *     if p_exon_num == 1 and r_exon_num > 1:             # <<<<<<<<<<<<<<
 *         if nucl_precision < 1 and nucl_overlap > 0:
 *             index = -1
 */
    goto __pyx_L5;
  }

  /* "Mikado/scales/contrast.pyx":98
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:             # <<<<<<<<<<<<<<
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 *             ccode = "n"
 */
  __pyx_t_11 = ((__pyx_v_p_exon_num > 1) != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_5 = __pyx_t_11;
    goto __pyx_L24_bool_binop_done;
  }
  __pyx_t_11 = ((__pyx_v_r_exon_num == 1) != 0);
  __pyx_t_5 = __pyx_t_11;
  __pyx_L24_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Mikado/scales/contrast.pyx":99
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference             # <<<<<<<<<<<<<<
 *             ccode = "n"
 *         elif nucl_overlap > 0:
 */
    __pyx_t_5 = ((__pyx_v_nucl_recall == 1.0) != 0);
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":100
 *     elif p_exon_num > 1 and r_exon_num == 1:
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 *             ccode = "n"             # <<<<<<<<<<<<<<
 *         elif nucl_overlap > 0:
 *             ccode = "G"  # Reverse generic overlap
 */
      __Pyx_INCREF(__pyx_n_u_n);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_n);

      /* "Mikado/scales/contrast.pyx":99
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference             # <<<<<<<<<<<<<<
 *             ccode = "n"
 *         elif nucl_overlap > 0:
 */
      goto __pyx_L26;
    }

    /* "Mikado/scales/contrast.pyx":101
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 *             ccode = "n"
 *         elif nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             ccode = "G"  # Reverse generic overlap
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:
 */
    __pyx_t_5 = ((__pyx_v_nucl_overlap > 0) != 0);
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":102
 *             ccode = "n"
 *         elif nucl_overlap > 0:
 *             ccode = "G"  # Reverse generic overlap             # <<<<<<<<<<<<<<
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:
 *             ccode = "ri"
 */
      __Pyx_INCREF(__pyx_n_u_G);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_G);

      /* "Mikado/scales/contrast.pyx":101
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 *             ccode = "n"
 *         elif nucl_overlap > 0:             # <<<<<<<<<<<<<<
 *             ccode = "G"  # Reverse generic overlap
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:
 */
      goto __pyx_L26;
    }

    /* "Mikado/scales/contrast.pyx":103
 *         elif nucl_overlap > 0:
 *             ccode = "G"  # Reverse generic overlap
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:             # <<<<<<<<<<<<<<
 *             ccode = "ri"
 *         else:
 */
    __pyx_t_5 = ((__pyx_f_6Mikado_9utilities_7overlap_c_overlap(__pyx_v_p_start, __pyx_v_p_end, __pyx_v_r_start, __pyx_v_r_end, 0, 1) > 0) != 0);
    if (__pyx_t_5) {

      /* "Mikado/scales/contrast.pyx":104
 *             ccode = "G"  # Reverse generic overlap
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:
 *             ccode = "ri"             # <<<<<<<<<<<<<<
 *         else:
 *             ccode = "p"
 */
      __Pyx_INCREF(__pyx_n_u_ri);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_ri);

      /* "Mikado/scales/contrast.pyx":103
 *         elif nucl_overlap > 0:
 *             ccode = "G"  # Reverse generic overlap
 *         elif c_overlap(p_start, p_end, r_start, r_end, 0, 1) > 0:             # <<<<<<<<<<<<<<
 *             ccode = "ri"
 *         else:
 */
      goto __pyx_L26;
    }

    /* "Mikado/scales/contrast.pyx":106
 *             ccode = "ri"
 *         else:
 *             ccode = "p"             # <<<<<<<<<<<<<<
 *     elif p_exon_num == r_exon_num == 1:
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_u_p);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_p);
    }
    __pyx_L26:;

    /* "Mikado/scales/contrast.pyx":98
 *         elif (nucl_recall == 0 and r_start < p_start < r_end):
 *             ccode = "i"  # Monoexonic fragment inside an intron
 *     elif p_exon_num > 1 and r_exon_num == 1:             # <<<<<<<<<<<<<<
 *         if nucl_recall == 1.0:  # The multiexonic is a complete extension of the monoexonic reference
 *             ccode = "n"
 */
    goto __pyx_L5;
  }

  /* "Mikado/scales/contrast.pyx":107
 *         else:
 *             ccode = "p"
 *     elif p_exon_num == r_exon_num == 1:             # <<<<<<<<<<<<<<
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:
 *             reference_exon = reference.exons[0]
 */
  __pyx_t_5 = (__pyx_v_p_exon_num == __pyx_v_r_exon_num);
  if (__pyx_t_5) {
    __pyx_t_5 = (__pyx_v_r_exon_num == 1);
  }
  __pyx_t_11 = (__pyx_t_5 != 0);
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":108
 *             ccode = "p"
 *     elif p_exon_num == r_exon_num == 1:
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:             # <<<<<<<<<<<<<<
 *             reference_exon = reference.exons[0]
 *             ccode = "_"
 */
    __pyx_t_5 = ((__pyx_v_nucl_f1 >= 0.95) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_11 = __pyx_t_5;
      goto __pyx_L28_bool_binop_done;
    }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_r_strand, __pyx_v_p_strand, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 108, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    __pyx_t_11 = __pyx_t_6;
    __pyx_L28_bool_binop_done:;
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":109
 *     elif p_exon_num == r_exon_num == 1:
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:
 *             reference_exon = reference.exons[0]             # <<<<<<<<<<<<<<
 *             ccode = "_"
 *         elif nucl_precision == 1:
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_reference_exon = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":110
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:
 *             reference_exon = reference.exons[0]
 *             ccode = "_"             # <<<<<<<<<<<<<<
 *         elif nucl_precision == 1:
 *             ccode = "c"  # contained
 */
      __Pyx_INCREF(__pyx_n_u__2);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u__2);

      /* "Mikado/scales/contrast.pyx":108
 *             ccode = "p"
 *     elif p_exon_num == r_exon_num == 1:
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:             # <<<<<<<<<<<<<<
 *             reference_exon = reference.exons[0]
 *             ccode = "_"
 */
      goto __pyx_L27;
    }

    /* "Mikado/scales/contrast.pyx":111
 *             reference_exon = reference.exons[0]
 *             ccode = "_"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"  # contained
 *         else:
 */
    __pyx_t_11 = ((__pyx_v_nucl_precision == 1.0) != 0);
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":112
 *             ccode = "_"
 *         elif nucl_precision == 1:
 *             ccode = "c"  # contained             # <<<<<<<<<<<<<<
 *         else:
 *             ccode = "m"  # just a generic exon overlap b/w two monoexonic transcripts
 */
      __Pyx_INCREF(__pyx_n_u_c);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_c);

      /* "Mikado/scales/contrast.pyx":111
 *             reference_exon = reference.exons[0]
 *             ccode = "_"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"  # contained
 *         else:
 */
      goto __pyx_L27;
    }

    /* "Mikado/scales/contrast.pyx":114
 *             ccode = "c"  # contained
 *         else:
 *             ccode = "m"  # just a generic exon overlap b/w two monoexonic transcripts             # <<<<<<<<<<<<<<
 * 
 *     return ccode
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_u_m);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_m);
    }
    __pyx_L27:;

    /* "Mikado/scales/contrast.pyx":107
 *         else:
 *             ccode = "p"
 *     elif p_exon_num == r_exon_num == 1:             # <<<<<<<<<<<<<<
 *         if nucl_f1 >= 0.95 and r_strand == p_strand:
 *             reference_exon = reference.exons[0]
 */
  }
  __pyx_L5:;

  /* "Mikado/scales/contrast.pyx":116
 *             ccode = "m"  # just a generic exon overlap b/w two monoexonic transcripts
 * 
 *     return ccode             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ccode);
  __pyx_r = __pyx_v_ccode;
  goto __pyx_L0;

  /* "Mikado/scales/contrast.pyx":16
 * @cython.profile(True)
 * @cython.boundscheck(False)
 * cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9]):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Mikado.scales.contrast.__assign_monoexonic_ccode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ccode);
  __Pyx_XDECREF(__pyx_v_r_strand);
  __Pyx_XDECREF(__pyx_v_p_strand);
  __Pyx_XDECREF(__pyx_v_intron);
  __Pyx_XDECREF(__pyx_v_reference_exon);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Mikado/scales/contrast.pyx":119
 * 
 * 
 * cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=0):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

static PyObject *__pyx_f_6Mikado_6scales_8contrast___assign_multiexonic_ccode(PyObject *__pyx_v_prediction, PyObject *__pyx_v_reference, long __pyx_v_nucl_overlap, double *__pyx_v_stats, struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode *__pyx_optional_args) {
  PyObject *__pyx_v_ccode = 0;
  CYTHON_UNUSED double __pyx_v_nucl_recall;
  double __pyx_v_nucl_precision;
  double __pyx_v_nucl_f1;
  CYTHON_UNUSED double __pyx_v_exon_recall;
  CYTHON_UNUSED double __pyx_v_exon_precision;
  CYTHON_UNUSED double __pyx_v_exon_f1;
  double __pyx_v_junction_recall;
  double __pyx_v_junction_precision;
  CYTHON_UNUSED double __pyx_v_junction_f1;
  CYTHON_UNUSED int __pyx_v_p_exon_num;
  CYTHON_UNUSED int __pyx_v_r_exon_num;
  long __pyx_v_p_start;
  long __pyx_v_p_end;
  long __pyx_v_r_start;
  long __pyx_v_r_end;
  PyObject *__pyx_v_r_splices = 0;
  PyObject *__pyx_v_p_splices = 0;
  PyObject *__pyx_v_r_introns = 0;
  PyObject *__pyx_v_p_introns = 0;
  CYTHON_UNUSED struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *__pyx_v_r_segtree = 0;
  CYTHON_UNUSED struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *__pyx_v_p_segtree = 0;
  long __pyx_v_intron_start;
  long __pyx_v_intron_end;
  int __pyx_v_start_in;
  long __pyx_v_min_splice;
  long __pyx_v_max_splice;
  long __pyx_v_splice;
  unsigned short __pyx_v_overlapping_introns;
  PyObject *__pyx_v_missed_introns = NULL;
  long __pyx_v_end_in;
  PyObject *__pyx_v_intron = NULL;
  PyObject *__pyx_v_ref_intron = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *(*__pyx_t_15)(PyObject *);
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  long __pyx_t_19;
  long __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__assign_multiexonic_ccode", 0);
  if (__pyx_optional_args) {
  }

  /* "Mikado/scales/contrast.pyx":152
 *         short unsigned int index
 * 
 *     r_splices, p_splices = reference.splices, prediction.splices             # <<<<<<<<<<<<<<
 *     r_introns, p_introns = reference.introns, prediction.introns
 *     r_segtree, p_segtree = reference.segmenttree, prediction.segmenttree
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_splices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_splices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_v_r_splices = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_p_splices = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Mikado/scales/contrast.pyx":153
 * 
 *     r_splices, p_splices = reference.splices, prediction.splices
 *     r_introns, p_introns = reference.introns, prediction.introns             # <<<<<<<<<<<<<<
 *     r_segtree, p_segtree = reference.segmenttree, prediction.segmenttree
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_introns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_introns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_v_r_introns = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_p_introns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":154
 *     r_splices, p_splices = reference.splices, prediction.splices
 *     r_introns, p_introns = reference.introns, prediction.introns
 *     r_segtree, p_segtree = reference.segmenttree, prediction.segmenttree             # <<<<<<<<<<<<<<
 * 
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_segmenttree); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree))))) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_segmenttree); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree))))) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_v_r_segtree = ((struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_p_segtree = ((struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Mikado/scales/contrast.pyx":156
 *     r_segtree, p_segtree = reference.segmenttree, prediction.segmenttree
 * 
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end             # <<<<<<<<<<<<<<
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_p_exon_num = __pyx_t_3;
  __pyx_v_p_start = __pyx_t_4;
  __pyx_v_p_end = __pyx_t_5;

  /* "Mikado/scales/contrast.pyx":157
 * 
 *     p_exon_num, p_start, p_end = prediction.exon_num, prediction.start, prediction.end
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end             # <<<<<<<<<<<<<<
 * 
 *     nucl_recall = stats[0]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_r_exon_num = __pyx_t_3;
  __pyx_v_r_start = __pyx_t_5;
  __pyx_v_r_end = __pyx_t_4;

  /* "Mikado/scales/contrast.pyx":159
 *     r_exon_num, r_start, r_end = reference.exon_num, reference.start, reference.end
 * 
 *     nucl_recall = stats[0]             # <<<<<<<<<<<<<<
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]
 */
  __pyx_v_nucl_recall = (__pyx_v_stats[0]);

  /* "Mikado/scales/contrast.pyx":160
 * 
 *     nucl_recall = stats[0]
 *     nucl_precision = stats[1]             # <<<<<<<<<<<<<<
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]
 */
  __pyx_v_nucl_precision = (__pyx_v_stats[1]);

  /* "Mikado/scales/contrast.pyx":161
 *     nucl_recall = stats[0]
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]             # <<<<<<<<<<<<<<
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]
 */
  __pyx_v_nucl_f1 = (__pyx_v_stats[2]);

  /* "Mikado/scales/contrast.pyx":162
 *     nucl_precision = stats[1]
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]             # <<<<<<<<<<<<<<
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]
 */
  __pyx_v_exon_recall = (__pyx_v_stats[3]);

  /* "Mikado/scales/contrast.pyx":163
 *     nucl_f1 = stats[2]
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]             # <<<<<<<<<<<<<<
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]
 */
  __pyx_v_exon_precision = (__pyx_v_stats[4]);

  /* "Mikado/scales/contrast.pyx":164
 *     exon_recall = stats[3]
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]             # <<<<<<<<<<<<<<
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]
 */
  __pyx_v_exon_f1 = (__pyx_v_stats[5]);

  /* "Mikado/scales/contrast.pyx":165
 *     exon_precision = stats[4]
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]             # <<<<<<<<<<<<<<
 *     junction_precision = stats[7]
 *     junction_f1 = stats[8]
 */
  __pyx_v_junction_recall = (__pyx_v_stats[6]);

  /* "Mikado/scales/contrast.pyx":166
 *     exon_f1 = stats[5]
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]             # <<<<<<<<<<<<<<
 *     junction_f1 = stats[8]
 * 
 */
  __pyx_v_junction_precision = (__pyx_v_stats[7]);

  /* "Mikado/scales/contrast.pyx":167
 *     junction_recall = stats[6]
 *     junction_precision = stats[7]
 *     junction_f1 = stats[8]             # <<<<<<<<<<<<<<
 * 
 *     ccode = ""
 */
  __pyx_v_junction_f1 = (__pyx_v_stats[8]);

  /* "Mikado/scales/contrast.pyx":169
 *     junction_f1 = stats[8]
 * 
 *     ccode = ""             # <<<<<<<<<<<<<<
 *     if junction_recall == 1 and junction_precision < 1:
 *         # Check if this is an extension
 */
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_v_ccode = __pyx_kp_u_;

  /* "Mikado/scales/contrast.pyx":170
 * 
 *     ccode = ""
 *     if junction_recall == 1 and junction_precision < 1:             # <<<<<<<<<<<<<<
 *         # Check if this is an extension
 *         # If one of the new splices is internal to the intron chain, it's a j
 */
  __pyx_t_7 = ((__pyx_v_junction_recall == 1.0) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = ((__pyx_v_junction_precision < 1.0) != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "Mikado/scales/contrast.pyx":174
 *         # If one of the new splices is internal to the intron chain, it's a j
 * 
 *         ccode = "n"             # <<<<<<<<<<<<<<
 *         max_splice = max(reference.splices)
 *         min_splice = min(reference.splices)
 */
    __Pyx_INCREF(__pyx_n_u_n);
    __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_n);

    /* "Mikado/scales/contrast.pyx":175
 * 
 *         ccode = "n"
 *         max_splice = max(reference.splices)             # <<<<<<<<<<<<<<
 *         min_splice = min(reference.splices)
 *         for splice in p_splices:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_splices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_max_splice = __pyx_t_4;

    /* "Mikado/scales/contrast.pyx":176
 *         ccode = "n"
 *         max_splice = max(reference.splices)
 *         min_splice = min(reference.splices)             # <<<<<<<<<<<<<<
 *         for splice in p_splices:
 *             if splice in r_splices:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_splices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_min_splice = __pyx_t_4;

    /* "Mikado/scales/contrast.pyx":177
 *         max_splice = max(reference.splices)
 *         min_splice = min(reference.splices)
 *         for splice in p_splices:             # <<<<<<<<<<<<<<
 *             if splice in r_splices:
 *                 continue
 */
    __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_set_iterator(__pyx_v_p_splices, 1, (&__pyx_t_9), (&__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_10 = __Pyx_set_iter_next(__pyx_t_2, __pyx_t_9, &__pyx_t_8, &__pyx_t_1, __pyx_t_3);
      if (unlikely(__pyx_t_10 == 0)) break;
      if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_splice = __pyx_t_4;

      /* "Mikado/scales/contrast.pyx":178
 *         min_splice = min(reference.splices)
 *         for splice in p_splices:
 *             if splice in r_splices:             # <<<<<<<<<<<<<<
 *                 continue
 *             if min_splice < splice < max_splice:
 */
      __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_splice); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__pyx_v_r_splices == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 178, __pyx_L1_error)
      }
      __pyx_t_6 = (__Pyx_PySet_ContainsTF(__pyx_t_1, __pyx_v_r_splices, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = (__pyx_t_6 != 0);
      if (__pyx_t_7) {

        /* "Mikado/scales/contrast.pyx":179
 *         for splice in p_splices:
 *             if splice in r_splices:
 *                 continue             # <<<<<<<<<<<<<<
 *             if min_splice < splice < max_splice:
 *                 ccode = "j"
 */
        goto __pyx_L6_continue;

        /* "Mikado/scales/contrast.pyx":178
 *         min_splice = min(reference.splices)
 *         for splice in p_splices:
 *             if splice in r_splices:             # <<<<<<<<<<<<<<
 *                 continue
 *             if min_splice < splice < max_splice:
 */
      }

      /* "Mikado/scales/contrast.pyx":180
 *             if splice in r_splices:
 *                 continue
 *             if min_splice < splice < max_splice:             # <<<<<<<<<<<<<<
 *                 ccode = "j"
 *                 break
 */
      __pyx_t_7 = (__pyx_v_min_splice < __pyx_v_splice);
      if (__pyx_t_7) {
        __pyx_t_7 = (__pyx_v_splice < __pyx_v_max_splice);
      }
      __pyx_t_6 = (__pyx_t_7 != 0);
      if (__pyx_t_6) {

        /* "Mikado/scales/contrast.pyx":181
 *                 continue
 *             if min_splice < splice < max_splice:
 *                 ccode = "j"             # <<<<<<<<<<<<<<
 *                 break
 *             elif r_start < splice < r_end:
 */
        __Pyx_INCREF(__pyx_n_u_j);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_j);

        /* "Mikado/scales/contrast.pyx":182
 *             if min_splice < splice < max_splice:
 *                 ccode = "j"
 *                 break             # <<<<<<<<<<<<<<
 *             elif r_start < splice < r_end:
 *                 ccode = "J"
 */
        goto __pyx_L7_break;

        /* "Mikado/scales/contrast.pyx":180
 *             if splice in r_splices:
 *                 continue
 *             if min_splice < splice < max_splice:             # <<<<<<<<<<<<<<
 *                 ccode = "j"
 *                 break
 */
      }

      /* "Mikado/scales/contrast.pyx":183
 *                 ccode = "j"
 *                 break
 *             elif r_start < splice < r_end:             # <<<<<<<<<<<<<<
 *                 ccode = "J"
 * 
 */
      __pyx_t_6 = (__pyx_v_r_start < __pyx_v_splice);
      if (__pyx_t_6) {
        __pyx_t_6 = (__pyx_v_splice < __pyx_v_r_end);
      }
      __pyx_t_7 = (__pyx_t_6 != 0);
      if (__pyx_t_7) {

        /* "Mikado/scales/contrast.pyx":184
 *                 break
 *             elif r_start < splice < r_end:
 *                 ccode = "J"             # <<<<<<<<<<<<<<
 * 
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:
 */
        __Pyx_INCREF(__pyx_n_u_J);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_J);

        /* "Mikado/scales/contrast.pyx":183
 *                 ccode = "j"
 *                 break
 *             elif r_start < splice < r_end:             # <<<<<<<<<<<<<<
 *                 ccode = "J"
 * 
 */
      }
      __pyx_L6_continue:;
    }
    __pyx_L7_break:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Mikado/scales/contrast.pyx":170
 * 
 *     ccode = ""
 *     if junction_recall == 1 and junction_precision < 1:             # <<<<<<<<<<<<<<
 *         # Check if this is an extension
 *         # If one of the new splices is internal to the intron chain, it's a j
 */
    goto __pyx_L3;
  }

  /* "Mikado/scales/contrast.pyx":186
 *                 ccode = "J"
 * 
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:             # <<<<<<<<<<<<<<
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:
 */
  __pyx_t_6 = (0.0 < __pyx_v_junction_recall);
  if (__pyx_t_6) {
    __pyx_t_6 = (__pyx_v_junction_recall < 1.0);
  }
  __pyx_t_11 = (__pyx_t_6 != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_7 = __pyx_t_11;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_11 = (0.0 < __pyx_v_junction_precision);
  if (__pyx_t_11) {
    __pyx_t_11 = (__pyx_v_junction_precision < 1.0);
  }
  __pyx_t_6 = (__pyx_t_11 != 0);
  __pyx_t_7 = __pyx_t_6;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_7) {

    /* "Mikado/scales/contrast.pyx":187
 * 
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:
 *         ccode = "j"             # <<<<<<<<<<<<<<
 *     elif junction_precision == 1 and junction_recall < 1:
 *         if set.difference(p_introns, r_introns):
 */
    __Pyx_INCREF(__pyx_n_u_j);
    __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_j);

    /* "Mikado/scales/contrast.pyx":186
 *                 ccode = "J"
 * 
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:             # <<<<<<<<<<<<<<
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:
 */
    goto __pyx_L3;
  }

  /* "Mikado/scales/contrast.pyx":188
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:             # <<<<<<<<<<<<<<
 *         if set.difference(p_introns, r_introns):
 *             ccode = "j"
 */
  __pyx_t_6 = ((__pyx_v_junction_precision == 1.0) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_7 = __pyx_t_6;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_junction_recall < 1.0) != 0);
  __pyx_t_7 = __pyx_t_6;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_7) {

    /* "Mikado/scales/contrast.pyx":189
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:
 *         if set.difference(p_introns, r_introns):             # <<<<<<<<<<<<<<
 *             ccode = "j"
 *         elif nucl_precision == 1:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_difference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = NULL;
    __pyx_t_3 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_3 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_p_introns, __pyx_v_r_introns};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_3, 2+__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_p_introns, __pyx_v_r_introns};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_3, 2+__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_12) {
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12); __pyx_t_12 = NULL;
      }
      __Pyx_INCREF(__pyx_v_p_introns);
      __Pyx_GIVEREF(__pyx_v_p_introns);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_3, __pyx_v_p_introns);
      __Pyx_INCREF(__pyx_v_r_introns);
      __Pyx_GIVEREF(__pyx_v_r_introns);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_3, __pyx_v_r_introns);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {

      /* "Mikado/scales/contrast.pyx":190
 *     elif junction_precision == 1 and junction_recall < 1:
 *         if set.difference(p_introns, r_introns):
 *             ccode = "j"             # <<<<<<<<<<<<<<
 *         elif nucl_precision == 1:
 *             ccode = "c"
 */
      __Pyx_INCREF(__pyx_n_u_j);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_j);

      /* "Mikado/scales/contrast.pyx":189
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:
 *         if set.difference(p_introns, r_introns):             # <<<<<<<<<<<<<<
 *             ccode = "j"
 *         elif nucl_precision == 1:
 */
      goto __pyx_L14;
    }

    /* "Mikado/scales/contrast.pyx":191
 *         if set.difference(p_introns, r_introns):
 *             ccode = "j"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"
 *         else:
 */
    __pyx_t_7 = ((__pyx_v_nucl_precision == 1.0) != 0);
    if (__pyx_t_7) {

      /* "Mikado/scales/contrast.pyx":192
 *             ccode = "j"
 *         elif nucl_precision == 1:
 *             ccode = "c"             # <<<<<<<<<<<<<<
 *         else:
 *             missed_introns = r_introns - p_introns
 */
      __Pyx_INCREF(__pyx_n_u_c);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_c);

      /* "Mikado/scales/contrast.pyx":191
 *         if set.difference(p_introns, r_introns):
 *             ccode = "j"
 *         elif nucl_precision == 1:             # <<<<<<<<<<<<<<
 *             ccode = "c"
 *         else:
 */
      goto __pyx_L14;
    }

    /* "Mikado/scales/contrast.pyx":194
 *             ccode = "c"
 *         else:
 *             missed_introns = r_introns - p_introns             # <<<<<<<<<<<<<<
 *             min_splice, max_splice = min(prediction.splices), max(prediction.splices)
 *             start_in = 0
 */
    /*else*/ {
      __pyx_t_2 = PyNumber_Subtract(__pyx_v_r_introns, __pyx_v_p_introns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_missed_introns = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "Mikado/scales/contrast.pyx":195
 *         else:
 *             missed_introns = r_introns - p_introns
 *             min_splice, max_splice = min(prediction.splices), max(prediction.splices)             # <<<<<<<<<<<<<<
 *             start_in = 0
 *             end_in = 0
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_splices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_splices); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_min_splice = __pyx_t_4;
      __pyx_v_max_splice = __pyx_t_5;

      /* "Mikado/scales/contrast.pyx":196
 *             missed_introns = r_introns - p_introns
 *             min_splice, max_splice = min(prediction.splices), max(prediction.splices)
 *             start_in = 0             # <<<<<<<<<<<<<<
 *             end_in = 0
 *             for intron in missed_introns:
 */
      __pyx_v_start_in = 0;

      /* "Mikado/scales/contrast.pyx":197
 *             min_splice, max_splice = min(prediction.splices), max(prediction.splices)
 *             start_in = 0
 *             end_in = 0             # <<<<<<<<<<<<<<
 *             for intron in missed_introns:
 *                 intron_start, intron_end = intron
 */
      __pyx_v_end_in = 0;

      /* "Mikado/scales/contrast.pyx":198
 *             start_in = 0
 *             end_in = 0
 *             for intron in missed_introns:             # <<<<<<<<<<<<<<
 *                 intron_start, intron_end = intron
 *                 if p_start < intron_start and intron_end < min_splice:
 */
      if (likely(PyList_CheckExact(__pyx_v_missed_introns)) || PyTuple_CheckExact(__pyx_v_missed_introns)) {
        __pyx_t_2 = __pyx_v_missed_introns; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_missed_introns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 198, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 198, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_14(__pyx_t_2);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 198, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_XDECREF_SET(__pyx_v_intron, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Mikado/scales/contrast.pyx":199
 *             end_in = 0
 *             for intron in missed_introns:
 *                 intron_start, intron_end = intron             # <<<<<<<<<<<<<<
 *                 if p_start < intron_start and intron_end < min_splice:
 *                     start_in = 1
 */
        if ((likely(PyTuple_CheckExact(__pyx_v_intron))) || (PyList_CheckExact(__pyx_v_intron))) {
          PyObject* sequence = __pyx_v_intron;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 199, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_13);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_13 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 199, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_12 = PyObject_GetIter(__pyx_v_intron); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 199, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_15 = Py_TYPE(__pyx_t_12)->tp_iternext;
          index = 0; __pyx_t_1 = __pyx_t_15(__pyx_t_12); if (unlikely(!__pyx_t_1)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_13 = __pyx_t_15(__pyx_t_12); if (unlikely(!__pyx_t_13)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_13);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_12), 2) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
          __pyx_t_15 = NULL;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L18_unpacking_done;
          __pyx_L17_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_15 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 199, __pyx_L1_error)
          __pyx_L18_unpacking_done:;
        }
        __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_13); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_v_intron_start = __pyx_t_5;
        __pyx_v_intron_end = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":200
 *             for intron in missed_introns:
 *                 intron_start, intron_end = intron
 *                 if p_start < intron_start and intron_end < min_splice:             # <<<<<<<<<<<<<<
 *                     start_in = 1
 *                     break
 */
        __pyx_t_6 = ((__pyx_v_p_start < __pyx_v_intron_start) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_7 = __pyx_t_6;
          goto __pyx_L20_bool_binop_done;
        }
        __pyx_t_6 = ((__pyx_v_intron_end < __pyx_v_min_splice) != 0);
        __pyx_t_7 = __pyx_t_6;
        __pyx_L20_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Mikado/scales/contrast.pyx":201
 *                 intron_start, intron_end = intron
 *                 if p_start < intron_start and intron_end < min_splice:
 *                     start_in = 1             # <<<<<<<<<<<<<<
 *                     break
 *                 if p_end > intron_end and intron_start > max_splice:
 */
          __pyx_v_start_in = 1;

          /* "Mikado/scales/contrast.pyx":202
 *                 if p_start < intron_start and intron_end < min_splice:
 *                     start_in = 1
 *                     break             # <<<<<<<<<<<<<<
 *                 if p_end > intron_end and intron_start > max_splice:
 *                     end_in = 1
 */
          goto __pyx_L16_break;

          /* "Mikado/scales/contrast.pyx":200
 *             for intron in missed_introns:
 *                 intron_start, intron_end = intron
 *                 if p_start < intron_start and intron_end < min_splice:             # <<<<<<<<<<<<<<
 *                     start_in = 1
 *                     break
 */
        }

        /* "Mikado/scales/contrast.pyx":203
 *                     start_in = 1
 *                     break
 *                 if p_end > intron_end and intron_start > max_splice:             # <<<<<<<<<<<<<<
 *                     end_in = 1
 *                     break
 */
        __pyx_t_6 = ((__pyx_v_p_end > __pyx_v_intron_end) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_7 = __pyx_t_6;
          goto __pyx_L23_bool_binop_done;
        }
        __pyx_t_6 = ((__pyx_v_intron_start > __pyx_v_max_splice) != 0);
        __pyx_t_7 = __pyx_t_6;
        __pyx_L23_bool_binop_done:;
        if (__pyx_t_7) {

          /* "Mikado/scales/contrast.pyx":204
 *                     break
 *                 if p_end > intron_end and intron_start > max_splice:
 *                     end_in = 1             # <<<<<<<<<<<<<<
 *                     break
 *             if start_in or end_in:
 */
          __pyx_v_end_in = 1;

          /* "Mikado/scales/contrast.pyx":205
 *                 if p_end > intron_end and intron_start > max_splice:
 *                     end_in = 1
 *                     break             # <<<<<<<<<<<<<<
 *             if start_in or end_in:
 *                 ccode = "j"
 */
          goto __pyx_L16_break;

          /* "Mikado/scales/contrast.pyx":203
 *                     start_in = 1
 *                     break
 *                 if p_end > intron_end and intron_start > max_splice:             # <<<<<<<<<<<<<<
 *                     end_in = 1
 *                     break
 */
        }

        /* "Mikado/scales/contrast.pyx":198
 *             start_in = 0
 *             end_in = 0
 *             for intron in missed_introns:             # <<<<<<<<<<<<<<
 *                 intron_start, intron_end = intron
 *                 if p_start < intron_start and intron_end < min_splice:
 */
      }
      __pyx_L16_break:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Mikado/scales/contrast.pyx":206
 *                     end_in = 1
 *                     break
 *             if start_in or end_in:             # <<<<<<<<<<<<<<
 *                 ccode = "j"
 *             else:
 */
      __pyx_t_6 = (__pyx_v_start_in != 0);
      if (!__pyx_t_6) {
      } else {
        __pyx_t_7 = __pyx_t_6;
        goto __pyx_L26_bool_binop_done;
      }
      __pyx_t_6 = (__pyx_v_end_in != 0);
      __pyx_t_7 = __pyx_t_6;
      __pyx_L26_bool_binop_done:;
      if (__pyx_t_7) {

        /* "Mikado/scales/contrast.pyx":207
 *                     break
 *             if start_in or end_in:
 *                 ccode = "j"             # <<<<<<<<<<<<<<
 *             else:
 *                 ccode = "C"
 */
        __Pyx_INCREF(__pyx_n_u_j);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_j);

        /* "Mikado/scales/contrast.pyx":206
 *                     end_in = 1
 *                     break
 *             if start_in or end_in:             # <<<<<<<<<<<<<<
 *                 ccode = "j"
 *             else:
 */
        goto __pyx_L25;
      }

      /* "Mikado/scales/contrast.pyx":209
 *                 ccode = "j"
 *             else:
 *                 ccode = "C"             # <<<<<<<<<<<<<<
 * 
 *     elif junction_recall == 0 and junction_precision == 0:
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_n_u_C);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_C);
      }
      __pyx_L25:;
    }
    __pyx_L14:;

    /* "Mikado/scales/contrast.pyx":188
 *     elif 0 < junction_recall < 1 and 0 < junction_precision < 1:
 *         ccode = "j"
 *     elif junction_precision == 1 and junction_recall < 1:             # <<<<<<<<<<<<<<
 *         if set.difference(p_introns, r_introns):
 *             ccode = "j"
 */
    goto __pyx_L3;
  }

  /* "Mikado/scales/contrast.pyx":211
 *                 ccode = "C"
 * 
 *     elif junction_recall == 0 and junction_precision == 0:             # <<<<<<<<<<<<<<
 *         if nucl_f1 > 0:
 *             overlapping_introns = 0
 */
  __pyx_t_6 = ((__pyx_v_junction_recall == 0.0) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_7 = __pyx_t_6;
    goto __pyx_L28_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_junction_precision == 0.0) != 0);
  __pyx_t_7 = __pyx_t_6;
  __pyx_L28_bool_binop_done:;
  if (__pyx_t_7) {

    /* "Mikado/scales/contrast.pyx":212
 * 
 *     elif junction_recall == 0 and junction_precision == 0:
 *         if nucl_f1 > 0:             # <<<<<<<<<<<<<<
 *             overlapping_introns = 0
 *             ccode = "o"
 */
    __pyx_t_7 = ((__pyx_v_nucl_f1 > 0.0) != 0);
    if (__pyx_t_7) {

      /* "Mikado/scales/contrast.pyx":213
 *     elif junction_recall == 0 and junction_precision == 0:
 *         if nucl_f1 > 0:
 *             overlapping_introns = 0             # <<<<<<<<<<<<<<
 *             ccode = "o"
 *             for intron in p_introns:
 */
      __pyx_v_overlapping_introns = 0;

      /* "Mikado/scales/contrast.pyx":214
 *         if nucl_f1 > 0:
 *             overlapping_introns = 0
 *             ccode = "o"             # <<<<<<<<<<<<<<
 *             for intron in p_introns:
 *                 for ref_intron in r_introns:
 */
      __Pyx_INCREF(__pyx_n_u_o);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_o);

      /* "Mikado/scales/contrast.pyx":215
 *             overlapping_introns = 0
 *             ccode = "o"
 *             for intron in p_introns:             # <<<<<<<<<<<<<<
 *                 for ref_intron in r_introns:
 *                     if c_overlap(ref_intron[0], ref_intron[1],
 */
      __pyx_t_9 = 0;
      __pyx_t_13 = __Pyx_set_iterator(__pyx_v_p_introns, 1, (&__pyx_t_8), (&__pyx_t_3)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_XDECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_13;
      __pyx_t_13 = 0;
      while (1) {
        __pyx_t_10 = __Pyx_set_iter_next(__pyx_t_2, __pyx_t_8, &__pyx_t_9, &__pyx_t_13, __pyx_t_3);
        if (unlikely(__pyx_t_10 == 0)) break;
        if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_XDECREF_SET(__pyx_v_intron, __pyx_t_13);
        __pyx_t_13 = 0;

        /* "Mikado/scales/contrast.pyx":216
 *             ccode = "o"
 *             for intron in p_introns:
 *                 for ref_intron in r_introns:             # <<<<<<<<<<<<<<
 *                     if c_overlap(ref_intron[0], ref_intron[1],
 *                                  intron[0], intron[1],
 */
        __pyx_t_16 = 0;
        __pyx_t_1 = __Pyx_set_iterator(__pyx_v_r_introns, 1, (&__pyx_t_17), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_13);
        __pyx_t_13 = __pyx_t_1;
        __pyx_t_1 = 0;
        while (1) {
          __pyx_t_18 = __Pyx_set_iter_next(__pyx_t_13, __pyx_t_17, &__pyx_t_16, &__pyx_t_1, __pyx_t_10);
          if (unlikely(__pyx_t_18 == 0)) break;
          if (unlikely(__pyx_t_18 == -1)) __PYX_ERR(0, 216, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_ref_intron, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "Mikado/scales/contrast.pyx":217
 *             for intron in p_introns:
 *                 for ref_intron in r_introns:
 *                     if c_overlap(ref_intron[0], ref_intron[1],             # <<<<<<<<<<<<<<
 *                                  intron[0], intron[1],
 *                                  0, 1) > 0:
 */
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_ref_intron, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_4 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 217, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_ref_intron, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 217, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Mikado/scales/contrast.pyx":218
 *                 for ref_intron in r_introns:
 *                     if c_overlap(ref_intron[0], ref_intron[1],
 *                                  intron[0], intron[1],             # <<<<<<<<<<<<<<
 *                                  0, 1) > 0:
 *                         # corr_exons.append((ref_index, pred_index))
 */
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_intron, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_19 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_19 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_intron, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Mikado/scales/contrast.pyx":219
 *                     if c_overlap(ref_intron[0], ref_intron[1],
 *                                  intron[0], intron[1],
 *                                  0, 1) > 0:             # <<<<<<<<<<<<<<
 *                         # corr_exons.append((ref_index, pred_index))
 *                         overlapping_introns += 1
 */
          __pyx_t_7 = ((__pyx_f_6Mikado_9utilities_7overlap_c_overlap(__pyx_t_4, __pyx_t_5, __pyx_t_19, __pyx_t_20, 0, 1) > 0) != 0);

          /* "Mikado/scales/contrast.pyx":217
 *             for intron in p_introns:
 *                 for ref_intron in r_introns:
 *                     if c_overlap(ref_intron[0], ref_intron[1],             # <<<<<<<<<<<<<<
 *                                  intron[0], intron[1],
 *                                  0, 1) > 0:
 */
          if (__pyx_t_7) {

            /* "Mikado/scales/contrast.pyx":221
 *                                  0, 1) > 0:
 *                         # corr_exons.append((ref_index, pred_index))
 *                         overlapping_introns += 1             # <<<<<<<<<<<<<<
 *                 if overlapping_introns >= 1:
 *                     ccode = "h"
 */
            __pyx_v_overlapping_introns = (__pyx_v_overlapping_introns + 1);

            /* "Mikado/scales/contrast.pyx":217
 *             for intron in p_introns:
 *                 for ref_intron in r_introns:
 *                     if c_overlap(ref_intron[0], ref_intron[1],             # <<<<<<<<<<<<<<
 *                                  intron[0], intron[1],
 *                                  0, 1) > 0:
 */
          }
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "Mikado/scales/contrast.pyx":222
 *                         # corr_exons.append((ref_index, pred_index))
 *                         overlapping_introns += 1
 *                 if overlapping_introns >= 1:             # <<<<<<<<<<<<<<
 *                     ccode = "h"
 *                     break
 */
        __pyx_t_7 = ((__pyx_v_overlapping_introns >= 1) != 0);
        if (__pyx_t_7) {

          /* "Mikado/scales/contrast.pyx":223
 *                         overlapping_introns += 1
 *                 if overlapping_introns >= 1:
 *                     ccode = "h"             # <<<<<<<<<<<<<<
 *                     break
 *         else:
 */
          __Pyx_INCREF(__pyx_n_u_h);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_h);

          /* "Mikado/scales/contrast.pyx":224
 *                 if overlapping_introns >= 1:
 *                     ccode = "h"
 *                     break             # <<<<<<<<<<<<<<
 *         else:
 *             if nucl_overlap == 0:
 */
          goto __pyx_L32_break;

          /* "Mikado/scales/contrast.pyx":222
 *                         # corr_exons.append((ref_index, pred_index))
 *                         overlapping_introns += 1
 *                 if overlapping_introns >= 1:             # <<<<<<<<<<<<<<
 *                     ccode = "h"
 *                     break
 */
        }
      }
      __pyx_L32_break:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Mikado/scales/contrast.pyx":212
 * 
 *     elif junction_recall == 0 and junction_precision == 0:
 *         if nucl_f1 > 0:             # <<<<<<<<<<<<<<
 *             overlapping_introns = 0
 *             ccode = "o"
 */
      goto __pyx_L30;
    }

    /* "Mikado/scales/contrast.pyx":226
 *                     break
 *         else:
 *             if nucl_overlap == 0:             # <<<<<<<<<<<<<<
 *                 # The only explanation for no nucleotide overlap
 *                 # and no junction overlap is that it is inside an intron
 */
    /*else*/ {
      __pyx_t_7 = ((__pyx_v_nucl_overlap == 0) != 0);
      if (__pyx_t_7) {

        /* "Mikado/scales/contrast.pyx":229
 *                 # The only explanation for no nucleotide overlap
 *                 # and no junction overlap is that it is inside an intron
 *                 if r_start < p_start < r_end:             # <<<<<<<<<<<<<<
 *                     ccode = "I"
 *                 elif p_start < r_start < p_end:
 */
        __pyx_t_7 = (__pyx_v_r_start < __pyx_v_p_start);
        if (__pyx_t_7) {
          __pyx_t_7 = (__pyx_v_p_start < __pyx_v_r_end);
        }
        __pyx_t_6 = (__pyx_t_7 != 0);
        if (__pyx_t_6) {

          /* "Mikado/scales/contrast.pyx":230
 *                 # and no junction overlap is that it is inside an intron
 *                 if r_start < p_start < r_end:
 *                     ccode = "I"             # <<<<<<<<<<<<<<
 *                 elif p_start < r_start < p_end:
 *                     ccode = "rI"  # reverse intron retention
 */
          __Pyx_INCREF(__pyx_n_u_I);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_I);

          /* "Mikado/scales/contrast.pyx":229
 *                 # The only explanation for no nucleotide overlap
 *                 # and no junction overlap is that it is inside an intron
 *                 if r_start < p_start < r_end:             # <<<<<<<<<<<<<<
 *                     ccode = "I"
 *                 elif p_start < r_start < p_end:
 */
          goto __pyx_L38;
        }

        /* "Mikado/scales/contrast.pyx":231
 *                 if r_start < p_start < r_end:
 *                     ccode = "I"
 *                 elif p_start < r_start < p_end:             # <<<<<<<<<<<<<<
 *                     ccode = "rI"  # reverse intron retention
 * 
 */
        __pyx_t_6 = (__pyx_v_p_start < __pyx_v_r_start);
        if (__pyx_t_6) {
          __pyx_t_6 = (__pyx_v_r_start < __pyx_v_p_end);
        }
        __pyx_t_7 = (__pyx_t_6 != 0);
        if (__pyx_t_7) {

          /* "Mikado/scales/contrast.pyx":232
 *                     ccode = "I"
 *                 elif p_start < r_start < p_end:
 *                     ccode = "rI"  # reverse intron retention             # <<<<<<<<<<<<<<
 * 
 *     return ccode
 */
          __Pyx_INCREF(__pyx_n_u_rI);
          __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_rI);

          /* "Mikado/scales/contrast.pyx":231
 *                 if r_start < p_start < r_end:
 *                     ccode = "I"
 *                 elif p_start < r_start < p_end:             # <<<<<<<<<<<<<<
 *                     ccode = "rI"  # reverse intron retention
 * 
 */
        }
        __pyx_L38:;

        /* "Mikado/scales/contrast.pyx":226
 *                     break
 *         else:
 *             if nucl_overlap == 0:             # <<<<<<<<<<<<<<
 *                 # The only explanation for no nucleotide overlap
 *                 # and no junction overlap is that it is inside an intron
 */
      }
    }
    __pyx_L30:;

    /* "Mikado/scales/contrast.pyx":211
 *                 ccode = "C"
 * 
 *     elif junction_recall == 0 and junction_precision == 0:             # <<<<<<<<<<<<<<
 *         if nucl_f1 > 0:
 *             overlapping_introns = 0
 */
  }
  __pyx_L3:;

  /* "Mikado/scales/contrast.pyx":234
 *                     ccode = "rI"  # reverse intron retention
 * 
 *     return ccode             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ccode);
  __pyx_r = __pyx_v_ccode;
  goto __pyx_L0;

  /* "Mikado/scales/contrast.pyx":119
 * 
 * 
 * cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=0):             # <<<<<<<<<<<<<<
 * 
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Mikado.scales.contrast.__assign_multiexonic_ccode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ccode);
  __Pyx_XDECREF(__pyx_v_r_splices);
  __Pyx_XDECREF(__pyx_v_p_splices);
  __Pyx_XDECREF(__pyx_v_r_introns);
  __Pyx_XDECREF(__pyx_v_p_introns);
  __Pyx_XDECREF((PyObject *)__pyx_v_r_segtree);
  __Pyx_XDECREF((PyObject *)__pyx_v_p_segtree);
  __Pyx_XDECREF(__pyx_v_missed_introns);
  __Pyx_XDECREF(__pyx_v_intron);
  __Pyx_XDECREF(__pyx_v_ref_intron);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Mikado/scales/contrast.pyx":239
 * @cython.profile(True)
 * @cython.cdivision(True)
 * cpdef tuple compare(prediction, reference, bint lenient=False, bint strict_strandedness=False, int fuzzymatch=0):             # <<<<<<<<<<<<<<
 * 
 *     """Cython function to compare two transcripts and determine a ccode.
 */

static PyObject *__pyx_pw_6Mikado_6scales_8contrast_1compare(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6Mikado_6scales_8contrast_compare(PyObject *__pyx_v_prediction, PyObject *__pyx_v_reference, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_6scales_8contrast_compare *__pyx_optional_args) {
  int __pyx_v_lenient = ((int)0);
  int __pyx_v_strict_strandedness = ((int)0);
  int __pyx_v_fuzzymatch = ((int)0);
  long __pyx_v_nucl_overlap;
  long __pyx_v_distance;
  PyObject *__pyx_v___pred_exons = 0;
  PyObject *__pyx_v___ref_exons = 0;
  long __pyx_v_exon_a;
  long __pyx_v_exon_b;
  long __pyx_v_other_exon_a;
  long __pyx_v_other_exon_b;
  double __pyx_v_p_cdna_length;
  double __pyx_v_r_cdna_length;
  double __pyx_v_nucl_recall;
  double __pyx_v_nucl_precision;
  double __pyx_v_nucl_f1;
  double __pyx_v_exon_recall;
  double __pyx_v_exon_precision;
  double __pyx_v_exon_f1;
  double __pyx_v_junction_recall;
  double __pyx_v_junction_precision;
  double __pyx_v_junction_f1;
  struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *__pyx_v_r_tree = 0;
  int __pyx_v_r_splice_start;
  int __pyx_v_r_splice_end;
  long __pyx_v_r_exon_num;
  long __pyx_v_p_exon_num;
  long __pyx_v_p_start;
  long __pyx_v_p_end;
  long __pyx_v_r_start;
  long __pyx_v_r_end;
  double __pyx_v_recalled_exons;
  PyObject *__pyx_v_r_splices = 0;
  PyObject *__pyx_v_p_splices = 0;
  PyObject *__pyx_v_c_splices = 0;
  int __pyx_v_p_splice_start;
  int __pyx_v_p_splice_end;
  int __pyx_v_best;
  long __pyx_v_junction_overlap;
  double __pyx_v_len_rsplices;
  double __pyx_v_len_psplices;
  double __pyx_v_stats[9];
  PyObject *__pyx_v_r_strand = 0;
  PyObject *__pyx_v_p_strand = 0;
  PyObject *__pyx_v_exon = NULL;
  PyObject *__pyx_v_other_exon = NULL;
  PyObject *__pyx_v_reference_exon = NULL;
  PyObject *__pyx_v_p_intron = NULL;
  PyObject *__pyx_v_found = NULL;
  long __pyx_v_curr_distance;
  PyObject *__pyx_v_r_intron = NULL;
  PyObject *__pyx_v_ccode = NULL;
  PyObject *__pyx_v_location = NULL;
  PyObject *__pyx_v_ref_parent = NULL;
  PyObject *__pyx_v_pred_parent = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_7genexpr__pyx_v__ = NULL;
  PyObject *__pyx_8genexpr1__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  double __pyx_t_5;
  double __pyx_t_6;
  long __pyx_t_7;
  long __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  int __pyx_t_16;
  struct __pyx_opt_args_6Mikado_9utilities_12intervaltree_12IntervalTree_find __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  Py_ssize_t __pyx_t_24;
  struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode __pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compare", 0);
  __Pyx_TraceCall("compare", __pyx_f[0], 239, 0, __PYX_ERR(0, 239, __pyx_L1_error));
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_lenient = __pyx_optional_args->lenient;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_strict_strandedness = __pyx_optional_args->strict_strandedness;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_fuzzymatch = __pyx_optional_args->fuzzymatch;
        }
      }
    }
  }

  /* "Mikado/scales/contrast.pyx":288
 *     """
 * 
 *     prediction.finalize()             # <<<<<<<<<<<<<<
 *     reference.finalize()
 *     fuzzymatch = abs(fuzzymatch)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_finalize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":289
 * 
 *     prediction.finalize()
 *     reference.finalize()             # <<<<<<<<<<<<<<
 *     fuzzymatch = abs(fuzzymatch)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_finalize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":290
 *     prediction.finalize()
 *     reference.finalize()
 *     fuzzymatch = abs(fuzzymatch)             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
  __pyx_t_4 = abs(__pyx_v_fuzzymatch); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 290, __pyx_L1_error)
  __pyx_v_fuzzymatch = __pyx_t_4;

  /* "Mikado/scales/contrast.pyx":315
 *         str r_strand, p_strand
 * 
 *     p_cdna_length, r_cdna_length = prediction.cdna_length, reference.cdna_length             # <<<<<<<<<<<<<<
 *     p_exon_num, r_exon_num = prediction.exon_num, reference.exon_num
 *     p_start, p_end, r_start, r_end = (prediction.start, prediction.end,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_cdna_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_cdna_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_cdna_length = __pyx_t_5;
  __pyx_v_r_cdna_length = __pyx_t_6;

  /* "Mikado/scales/contrast.pyx":316
 * 
 *     p_cdna_length, r_cdna_length = prediction.cdna_length, reference.cdna_length
 *     p_exon_num, r_exon_num = prediction.exon_num, reference.exon_num             # <<<<<<<<<<<<<<
 *     p_start, p_end, r_start, r_end = (prediction.start, prediction.end,
 *                                       reference.start, reference.end)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exon_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_8 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_exon_num = __pyx_t_7;
  __pyx_v_r_exon_num = __pyx_t_8;

  /* "Mikado/scales/contrast.pyx":317
 *     p_cdna_length, r_cdna_length = prediction.cdna_length, reference.cdna_length
 *     p_exon_num, r_exon_num = prediction.exon_num, reference.exon_num
 *     p_start, p_end, r_start, r_end = (prediction.start, prediction.end,             # <<<<<<<<<<<<<<
 *                                       reference.start, reference.end)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_8 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":318
 *     p_exon_num, r_exon_num = prediction.exon_num, reference.exon_num
 *     p_start, p_end, r_start, r_end = (prediction.start, prediction.end,
 *                                       reference.start, reference.end)             # <<<<<<<<<<<<<<
 * 
 *     if reference.strand is None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_9 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_10 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p_start = __pyx_t_8;
  __pyx_v_p_end = __pyx_t_7;
  __pyx_v_r_start = __pyx_t_9;
  __pyx_v_r_end = __pyx_t_10;

  /* "Mikado/scales/contrast.pyx":320
 *                                       reference.start, reference.end)
 * 
 *     if reference.strand is None:             # <<<<<<<<<<<<<<
 *         r_strand = ""
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = (__pyx_t_11 != 0);
  if (__pyx_t_12) {

    /* "Mikado/scales/contrast.pyx":321
 * 
 *     if reference.strand is None:
 *         r_strand = ""             # <<<<<<<<<<<<<<
 *     else:
 *         r_strand = reference.strand
 */
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_v_r_strand = __pyx_kp_u_;

    /* "Mikado/scales/contrast.pyx":320
 *                                       reference.start, reference.end)
 * 
 *     if reference.strand is None:             # <<<<<<<<<<<<<<
 *         r_strand = ""
 *     else:
 */
    goto __pyx_L3;
  }

  /* "Mikado/scales/contrast.pyx":323
 *         r_strand = ""
 *     else:
 *         r_strand = reference.strand             # <<<<<<<<<<<<<<
 * 
 *     if prediction.strand is None:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 323, __pyx_L1_error)
    __pyx_v_r_strand = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Mikado/scales/contrast.pyx":325
 *         r_strand = reference.strand
 * 
 *     if prediction.strand is None:             # <<<<<<<<<<<<<<
 *         p_strand = ""
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_12 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = (__pyx_t_12 != 0);
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":326
 * 
 *     if prediction.strand is None:
 *         p_strand = ""             # <<<<<<<<<<<<<<
 *     else:
 *         p_strand = prediction.strand
 */
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_v_p_strand = __pyx_kp_u_;

    /* "Mikado/scales/contrast.pyx":325
 *         r_strand = reference.strand
 * 
 *     if prediction.strand is None:             # <<<<<<<<<<<<<<
 *         p_strand = ""
 *     else:
 */
    goto __pyx_L4;
  }

  /* "Mikado/scales/contrast.pyx":328
 *         p_strand = ""
 *     else:
 *         p_strand = prediction.strand             # <<<<<<<<<<<<<<
 * 
 *     nucl_overlap = 0
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 328, __pyx_L1_error)
    __pyx_v_p_strand = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "Mikado/scales/contrast.pyx":330
 *         p_strand = prediction.strand
 * 
 *     nucl_overlap = 0             # <<<<<<<<<<<<<<
 * 
 *     __pred_exons = set()
 */
  __pyx_v_nucl_overlap = 0;

  /* "Mikado/scales/contrast.pyx":332
 *     nucl_overlap = 0
 * 
 *     __pred_exons = set()             # <<<<<<<<<<<<<<
 *     __ref_exons = set()
 *     r_tree = reference.segmenttree
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v___pred_exons = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":333
 * 
 *     __pred_exons = set()
 *     __ref_exons = set()             # <<<<<<<<<<<<<<
 *     r_tree = reference.segmenttree
 * 
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v___ref_exons = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":334
 *     __pred_exons = set()
 *     __ref_exons = set()
 *     r_tree = reference.segmenttree             # <<<<<<<<<<<<<<
 * 
 *     for exon in prediction.exons:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_segmenttree); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree))))) __PYX_ERR(0, 334, __pyx_L1_error)
  __pyx_v_r_tree = ((struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":336
 *     r_tree = reference.segmenttree
 * 
 *     for exon in prediction.exons:             # <<<<<<<<<<<<<<
 *         exon_a, exon_b = exon[0], exon[1] + 1
 *         __pred_exons.add(exon)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_13 = 0;
    __pyx_t_14 = NULL;
  } else {
    __pyx_t_13 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 336, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_14)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_13); __Pyx_INCREF(__pyx_t_1); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_14(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 336, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_exon, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Mikado/scales/contrast.pyx":337
 * 
 *     for exon in prediction.exons:
 *         exon_a, exon_b = exon[0], exon[1] + 1             # <<<<<<<<<<<<<<
 *         __pred_exons.add(exon)
 *                     # self, int start, int end, bint strict=0, bint contained_check=0, int max_distance=0,
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_exon, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_10 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_exon, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_long(__pyx_t_3); if (unlikely((__pyx_t_9 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_exon_a = __pyx_t_10;
    __pyx_v_exon_b = __pyx_t_9;

    /* "Mikado/scales/contrast.pyx":338
 *     for exon in prediction.exons:
 *         exon_a, exon_b = exon[0], exon[1] + 1
 *         __pred_exons.add(exon)             # <<<<<<<<<<<<<<
 *                     # self, int start, int end, bint strict=0, bint contained_check=0, int max_distance=0,
 *             #              int num_intervals=1000, object value=None
 */
    __pyx_t_15 = PySet_Add(__pyx_v___pred_exons, __pyx_v_exon); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 338, __pyx_L1_error)

    /* "Mikado/scales/contrast.pyx":341
 *                     # self, int start, int end, bint strict=0, bint contained_check=0, int max_distance=0,
 *             #              int num_intervals=1000, object value=None
 *         for other_exon in r_tree.find(exon[0], exon[1] + 1, 0, 0, fuzzymatch, 1000, "exon"):             # <<<<<<<<<<<<<<
 *             other_exon_a, other_exon_b = other_exon[0], other_exon[1] + 1
 *             __ref_exons.add((other_exon[0], other_exon[1]))
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_exon, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_exon, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_17.__pyx_n = 5;
    __pyx_t_17.strict = 0;
    __pyx_t_17.contained_check = 0;
    __pyx_t_17.max_distance = __pyx_v_fuzzymatch;
    __pyx_t_17.n = 0x3E8;
    __pyx_t_17.value = __pyx_n_u_exon;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_v_r_tree->__pyx_vtab)->find(__pyx_v_r_tree, __pyx_t_4, __pyx_t_16, 0, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_18 = 0;
      __pyx_t_19 = NULL;
    } else {
      __pyx_t_18 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_19 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 341, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_19)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 341, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 341, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_19(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 341, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_other_exon, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":342
 *             #              int num_intervals=1000, object value=None
 *         for other_exon in r_tree.find(exon[0], exon[1] + 1, 0, 0, fuzzymatch, 1000, "exon"):
 *             other_exon_a, other_exon_b = other_exon[0], other_exon[1] + 1             # <<<<<<<<<<<<<<
 *             __ref_exons.add((other_exon[0], other_exon[1]))
 *             nucl_overlap += c_overlap(exon_a, exon_b,
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_other_exon, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_9 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_other_exon, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_20 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = __Pyx_PyInt_As_long(__pyx_t_20); if (unlikely((__pyx_t_10 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __pyx_v_other_exon_a = __pyx_t_9;
      __pyx_v_other_exon_b = __pyx_t_10;

      /* "Mikado/scales/contrast.pyx":343
 *         for other_exon in r_tree.find(exon[0], exon[1] + 1, 0, 0, fuzzymatch, 1000, "exon"):
 *             other_exon_a, other_exon_b = other_exon[0], other_exon[1] + 1
 *             __ref_exons.add((other_exon[0], other_exon[1]))             # <<<<<<<<<<<<<<
 *             nucl_overlap += c_overlap(exon_a, exon_b,
 *                                       other_exon_a, other_exon_b,
 */
      __pyx_t_20 = __Pyx_GetItemInt(__pyx_v_other_exon, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_other_exon, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      __Pyx_GIVEREF(__pyx_t_20);
      PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_20);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_1);
      __pyx_t_20 = 0;
      __pyx_t_1 = 0;
      __pyx_t_15 = PySet_Add(__pyx_v___ref_exons, __pyx_t_21); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 343, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

      /* "Mikado/scales/contrast.pyx":344
 *             other_exon_a, other_exon_b = other_exon[0], other_exon[1] + 1
 *             __ref_exons.add((other_exon[0], other_exon[1]))
 *             nucl_overlap += c_overlap(exon_a, exon_b,             # <<<<<<<<<<<<<<
 *                                       other_exon_a, other_exon_b,
 *                                       flank=0,
 */
      __pyx_v_nucl_overlap = (__pyx_v_nucl_overlap + __pyx_f_6Mikado_9utilities_7overlap_c_overlap(__pyx_v_exon_a, __pyx_v_exon_b, __pyx_v_other_exon_a, __pyx_v_other_exon_b, 0, 1));

      /* "Mikado/scales/contrast.pyx":341
 *                     # self, int start, int end, bint strict=0, bint contained_check=0, int max_distance=0,
 *             #              int num_intervals=1000, object value=None
 *         for other_exon in r_tree.find(exon[0], exon[1] + 1, 0, 0, fuzzymatch, 1000, "exon"):             # <<<<<<<<<<<<<<
 *             other_exon_a, other_exon_b = other_exon[0], other_exon[1] + 1
 *             __ref_exons.add((other_exon[0], other_exon[1]))
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Mikado/scales/contrast.pyx":336
 *     r_tree = reference.segmenttree
 * 
 *     for exon in prediction.exons:             # <<<<<<<<<<<<<<
 *         exon_a, exon_b = exon[0], exon[1] + 1
 *         __pred_exons.add(exon)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Mikado/scales/contrast.pyx":349
 *                                       positive=1)
 * 
 *     nucl_recall = nucl_overlap / r_cdna_length  # Sensitivity             # <<<<<<<<<<<<<<
 *     nucl_precision = nucl_overlap / p_cdna_length
 *     nucl_f1 = calc_f1(nucl_recall, nucl_precision)
 */
  __pyx_v_nucl_recall = (((double)__pyx_v_nucl_overlap) / __pyx_v_r_cdna_length);

  /* "Mikado/scales/contrast.pyx":350
 * 
 *     nucl_recall = nucl_overlap / r_cdna_length  # Sensitivity
 *     nucl_precision = nucl_overlap / p_cdna_length             # <<<<<<<<<<<<<<
 *     nucl_f1 = calc_f1(nucl_recall, nucl_precision)
 * 
 */
  __pyx_v_nucl_precision = (((double)__pyx_v_nucl_overlap) / __pyx_v_p_cdna_length);

  /* "Mikado/scales/contrast.pyx":351
 *     nucl_recall = nucl_overlap / r_cdna_length  # Sensitivity
 *     nucl_precision = nucl_overlap / p_cdna_length
 *     nucl_f1 = calc_f1(nucl_recall, nucl_precision)             # <<<<<<<<<<<<<<
 * 
 *     # Exon statistics
 */
  __pyx_v_nucl_f1 = __pyx_f_6Mikado_9utilities_2f1_calc_f1(__pyx_v_nucl_recall, __pyx_v_nucl_precision, 0);

  /* "Mikado/scales/contrast.pyx":354
 * 
 *     # Exon statistics
 *     if lenient is True:             # <<<<<<<<<<<<<<
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:
 *             # If both are multiexonic, consider only the internal boundary
 */
  __pyx_t_11 = ((__pyx_v_lenient == 1) != 0);
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":355
 *     # Exon statistics
 *     if lenient is True:
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:             # <<<<<<<<<<<<<<
 *             # If both are multiexonic, consider only the internal boundary
 *             if prediction.exons[0] in __pred_exons:
 */
    __pyx_t_13 = PySet_GET_SIZE(__pyx_v___pred_exons); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 355, __pyx_L1_error)
    __pyx_t_12 = ((__pyx_t_13 > 1) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_13 = PySet_GET_SIZE(__pyx_v___ref_exons); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 355, __pyx_L1_error)
    __pyx_t_12 = ((__pyx_t_13 > 1) != 0);
    __pyx_t_11 = __pyx_t_12;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":357
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:
 *             # If both are multiexonic, consider only the internal boundary
 *             if prediction.exons[0] in __pred_exons:             # <<<<<<<<<<<<<<
 *                 __pred_exons.remove(prediction.exons[0])
 *                 __pred_exons.add(prediction.exons[0][1])
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_t_3, __pyx_v___pred_exons, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Mikado/scales/contrast.pyx":358
 *             # If both are multiexonic, consider only the internal boundary
 *             if prediction.exons[0] in __pred_exons:
 *                 __pred_exons.remove(prediction.exons[0])             # <<<<<<<<<<<<<<
 *                 __pred_exons.add(prediction.exons[0][1])
 *             if prediction.exons[-1] in __pred_exons:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PySet_Remove(__pyx_v___pred_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":359
 *             if prediction.exons[0] in __pred_exons:
 *                 __pred_exons.remove(prediction.exons[0])
 *                 __pred_exons.add(prediction.exons[0][1])             # <<<<<<<<<<<<<<
 *             if prediction.exons[-1] in __pred_exons:
 *                 __pred_exons.remove(prediction.exons[-1])
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = PySet_Add(__pyx_v___pred_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 359, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":357
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:
 *             # If both are multiexonic, consider only the internal boundary
 *             if prediction.exons[0] in __pred_exons:             # <<<<<<<<<<<<<<
 *                 __pred_exons.remove(prediction.exons[0])
 *                 __pred_exons.add(prediction.exons[0][1])
 */
      }

      /* "Mikado/scales/contrast.pyx":360
 *                 __pred_exons.remove(prediction.exons[0])
 *                 __pred_exons.add(prediction.exons[0][1])
 *             if prediction.exons[-1] in __pred_exons:             # <<<<<<<<<<<<<<
 *                 __pred_exons.remove(prediction.exons[-1])
 *                 __pred_exons.add(prediction.exons[-1][0])
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = (__Pyx_PySet_ContainsTF(__pyx_t_3, __pyx_v___pred_exons, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 360, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = (__pyx_t_12 != 0);
      if (__pyx_t_11) {

        /* "Mikado/scales/contrast.pyx":361
 *                 __pred_exons.add(prediction.exons[0][1])
 *             if prediction.exons[-1] in __pred_exons:
 *                 __pred_exons.remove(prediction.exons[-1])             # <<<<<<<<<<<<<<
 *                 __pred_exons.add(prediction.exons[-1][0])
 *             if reference.exons[0] in __ref_exons:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PySet_Remove(__pyx_v___pred_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":362
 *             if prediction.exons[-1] in __pred_exons:
 *                 __pred_exons.remove(prediction.exons[-1])
 *                 __pred_exons.add(prediction.exons[-1][0])             # <<<<<<<<<<<<<<
 *             if reference.exons[0] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[0])
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = PySet_Add(__pyx_v___pred_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":360
 *                 __pred_exons.remove(prediction.exons[0])
 *                 __pred_exons.add(prediction.exons[0][1])
 *             if prediction.exons[-1] in __pred_exons:             # <<<<<<<<<<<<<<
 *                 __pred_exons.remove(prediction.exons[-1])
 *                 __pred_exons.add(prediction.exons[-1][0])
 */
      }

      /* "Mikado/scales/contrast.pyx":363
 *                 __pred_exons.remove(prediction.exons[-1])
 *                 __pred_exons.add(prediction.exons[-1][0])
 *             if reference.exons[0] in __ref_exons:             # <<<<<<<<<<<<<<
 *                 __ref_exons.remove(reference.exons[0])
 *                 __ref_exons.add(reference.exons[0][1])
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = (__Pyx_PySet_ContainsTF(__pyx_t_3, __pyx_v___ref_exons, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Mikado/scales/contrast.pyx":364
 *                 __pred_exons.add(prediction.exons[-1][0])
 *             if reference.exons[0] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[0])             # <<<<<<<<<<<<<<
 *                 __ref_exons.add(reference.exons[0][1])
 *             if reference.exons[-1] in __ref_exons:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PySet_Remove(__pyx_v___ref_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":365
 *             if reference.exons[0] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[0])
 *                 __ref_exons.add(reference.exons[0][1])             # <<<<<<<<<<<<<<
 *             if reference.exons[-1] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[-1])
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = PySet_Add(__pyx_v___ref_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":363
 *                 __pred_exons.remove(prediction.exons[-1])
 *                 __pred_exons.add(prediction.exons[-1][0])
 *             if reference.exons[0] in __ref_exons:             # <<<<<<<<<<<<<<
 *                 __ref_exons.remove(reference.exons[0])
 *                 __ref_exons.add(reference.exons[0][1])
 */
      }

      /* "Mikado/scales/contrast.pyx":366
 *                 __ref_exons.remove(reference.exons[0])
 *                 __ref_exons.add(reference.exons[0][1])
 *             if reference.exons[-1] in __ref_exons:             # <<<<<<<<<<<<<<
 *                 __ref_exons.remove(reference.exons[-1])
 *                 __ref_exons.add(reference.exons[-1][0])
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = (__Pyx_PySet_ContainsTF(__pyx_t_3, __pyx_v___ref_exons, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = (__pyx_t_12 != 0);
      if (__pyx_t_11) {

        /* "Mikado/scales/contrast.pyx":367
 *                 __ref_exons.add(reference.exons[0][1])
 *             if reference.exons[-1] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[-1])             # <<<<<<<<<<<<<<
 *                 __ref_exons.add(reference.exons[-1][0])
 *         elif len(__pred_exons) == len(__ref_exons) == 1 and nucl_f1 > 0.8:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = __Pyx_PySet_Remove(__pyx_v___ref_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 367, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":368
 *             if reference.exons[-1] in __ref_exons:
 *                 __ref_exons.remove(reference.exons[-1])
 *                 __ref_exons.add(reference.exons[-1][0])             # <<<<<<<<<<<<<<
 *         elif len(__pred_exons) == len(__ref_exons) == 1 and nucl_f1 > 0.8:
 *             # If both are monoexonic and nucleotide F1 is >= 0.8
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_15 = PySet_Add(__pyx_v___ref_exons, __pyx_t_2); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Mikado/scales/contrast.pyx":366
 *                 __ref_exons.remove(reference.exons[0])
 *                 __ref_exons.add(reference.exons[0][1])
 *             if reference.exons[-1] in __ref_exons:             # <<<<<<<<<<<<<<
 *                 __ref_exons.remove(reference.exons[-1])
 *                 __ref_exons.add(reference.exons[-1][0])
 */
      }

      /* "Mikado/scales/contrast.pyx":355
 *     # Exon statistics
 *     if lenient is True:
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:             # <<<<<<<<<<<<<<
 *             # If both are multiexonic, consider only the internal boundary
 *             if prediction.exons[0] in __pred_exons:
 */
      goto __pyx_L10;
    }

    /* "Mikado/scales/contrast.pyx":369
 *                 __ref_exons.remove(reference.exons[-1])
 *                 __ref_exons.add(reference.exons[-1][0])
 *         elif len(__pred_exons) == len(__ref_exons) == 1 and nucl_f1 > 0.8:             # <<<<<<<<<<<<<<
 *             # If both are monoexonic and nucleotide F1 is >= 0.8
 *             __ref_exons = __pred_exons.copy()
 */
    __pyx_t_13 = PySet_GET_SIZE(__pyx_v___pred_exons); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 369, __pyx_L1_error)
    __pyx_t_18 = PySet_GET_SIZE(__pyx_v___ref_exons); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 369, __pyx_L1_error)
    __pyx_t_12 = (__pyx_t_13 == __pyx_t_18);
    if (__pyx_t_12) {
      __pyx_t_12 = (__pyx_t_18 == 1);
    }
    __pyx_t_22 = (__pyx_t_12 != 0);
    if (__pyx_t_22) {
    } else {
      __pyx_t_11 = __pyx_t_22;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_22 = ((__pyx_v_nucl_f1 > 0.8) != 0);
    __pyx_t_11 = __pyx_t_22;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":371
 *         elif len(__pred_exons) == len(__ref_exons) == 1 and nucl_f1 > 0.8:
 *             # If both are monoexonic and nucleotide F1 is >= 0.8
 *             __ref_exons = __pred_exons.copy()             # <<<<<<<<<<<<<<
 * 
 *     recalled_exons = len(set.intersection(__pred_exons, __ref_exons))
 */
      __pyx_t_2 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PySet_Type_copy, __pyx_v___pred_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 371, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v___ref_exons, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "Mikado/scales/contrast.pyx":369
 *                 __ref_exons.remove(reference.exons[-1])
 *                 __ref_exons.add(reference.exons[-1][0])
 *         elif len(__pred_exons) == len(__ref_exons) == 1 and nucl_f1 > 0.8:             # <<<<<<<<<<<<<<
 *             # If both are monoexonic and nucleotide F1 is >= 0.8
 *             __ref_exons = __pred_exons.copy()
 */
    }
    __pyx_L10:;

    /* "Mikado/scales/contrast.pyx":354
 * 
 *     # Exon statistics
 *     if lenient is True:             # <<<<<<<<<<<<<<
 *         if len(__pred_exons) > 1 and len(__ref_exons) > 1:
 *             # If both are multiexonic, consider only the internal boundary
 */
  }

  /* "Mikado/scales/contrast.pyx":373
 *             __ref_exons = __pred_exons.copy()
 * 
 *     recalled_exons = len(set.intersection(__pred_exons, __ref_exons))             # <<<<<<<<<<<<<<
 *     exon_recall = recalled_exons / r_exon_num
 *     exon_precision = recalled_exons / p_exon_num
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_intersection); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_21 = NULL;
  __pyx_t_16 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_21)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_21);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_16 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_v___pred_exons, __pyx_v___ref_exons};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_v___pred_exons, __pyx_v___ref_exons};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_21) {
      __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_21); __pyx_t_21 = NULL;
    }
    __Pyx_INCREF(__pyx_v___pred_exons);
    __Pyx_GIVEREF(__pyx_v___pred_exons);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_16, __pyx_v___pred_exons);
    __Pyx_INCREF(__pyx_v___ref_exons);
    __Pyx_GIVEREF(__pyx_v___ref_exons);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_16, __pyx_v___ref_exons);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_18 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_recalled_exons = __pyx_t_18;

  /* "Mikado/scales/contrast.pyx":374
 * 
 *     recalled_exons = len(set.intersection(__pred_exons, __ref_exons))
 *     exon_recall = recalled_exons / r_exon_num             # <<<<<<<<<<<<<<
 *     exon_precision = recalled_exons / p_exon_num
 *     exon_f1 = calc_f1(exon_recall, exon_precision)
 */
  __pyx_v_exon_recall = (__pyx_v_recalled_exons / ((double)__pyx_v_r_exon_num));

  /* "Mikado/scales/contrast.pyx":375
 *     recalled_exons = len(set.intersection(__pred_exons, __ref_exons))
 *     exon_recall = recalled_exons / r_exon_num
 *     exon_precision = recalled_exons / p_exon_num             # <<<<<<<<<<<<<<
 *     exon_f1 = calc_f1(exon_recall, exon_precision)
 * 
 */
  __pyx_v_exon_precision = (__pyx_v_recalled_exons / ((double)__pyx_v_p_exon_num));

  /* "Mikado/scales/contrast.pyx":376
 *     exon_recall = recalled_exons / r_exon_num
 *     exon_precision = recalled_exons / p_exon_num
 *     exon_f1 = calc_f1(exon_recall, exon_precision)             # <<<<<<<<<<<<<<
 * 
 *     reference_exon = None
 */
  __pyx_v_exon_f1 = __pyx_f_6Mikado_9utilities_2f1_calc_f1(__pyx_v_exon_recall, __pyx_v_exon_precision, 0);

  /* "Mikado/scales/contrast.pyx":378
 *     exon_f1 = calc_f1(exon_recall, exon_precision)
 * 
 *     reference_exon = None             # <<<<<<<<<<<<<<
 * 
 *     # Both multiexonic
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_reference_exon = Py_None;

  /* "Mikado/scales/contrast.pyx":381
 * 
 *     # Both multiexonic
 *     junction_overlap = 0             # <<<<<<<<<<<<<<
 *     c_splices = set()
 *     if p_exon_num > 1 and r_exon_num > 1:
 */
  __pyx_v_junction_overlap = 0;

  /* "Mikado/scales/contrast.pyx":382
 *     # Both multiexonic
 *     junction_overlap = 0
 *     c_splices = set()             # <<<<<<<<<<<<<<
 *     if p_exon_num > 1 and r_exon_num > 1:
 *         r_splices, p_splices = reference.splices, prediction.splices
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_c_splices = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Mikado/scales/contrast.pyx":383
 *     junction_overlap = 0
 *     c_splices = set()
 *     if p_exon_num > 1 and r_exon_num > 1:             # <<<<<<<<<<<<<<
 *         r_splices, p_splices = reference.splices, prediction.splices
 *         for p_intron in prediction.introns:
 */
  __pyx_t_22 = ((__pyx_v_p_exon_num > 1) != 0);
  if (__pyx_t_22) {
  } else {
    __pyx_t_11 = __pyx_t_22;
    goto __pyx_L20_bool_binop_done;
  }
  __pyx_t_22 = ((__pyx_v_r_exon_num > 1) != 0);
  __pyx_t_11 = __pyx_t_22;
  __pyx_L20_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":384
 *     c_splices = set()
 *     if p_exon_num > 1 and r_exon_num > 1:
 *         r_splices, p_splices = reference.splices, prediction.splices             # <<<<<<<<<<<<<<
 *         for p_intron in prediction.introns:
 *             p_splice_start, p_splice_end = p_intron
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_splices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PySet_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 384, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_splices); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PySet_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 384, __pyx_L1_error)
    __pyx_v_r_splices = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_p_splices = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Mikado/scales/contrast.pyx":385
 *     if p_exon_num > 1 and r_exon_num > 1:
 *         r_splices, p_splices = reference.splices, prediction.splices
 *         for p_intron in prediction.introns:             # <<<<<<<<<<<<<<
 *             p_splice_start, p_splice_end = p_intron
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_introns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_18 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_18 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 385, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_18); __Pyx_INCREF(__pyx_t_3); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 385, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_18); __Pyx_INCREF(__pyx_t_3); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 385, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_14(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 385, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_p_intron, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Mikado/scales/contrast.pyx":386
 *         r_splices, p_splices = reference.splices, prediction.splices
 *         for p_intron in prediction.introns:
 *             p_splice_start, p_splice_end = p_intron             # <<<<<<<<<<<<<<
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")
 *             curr_distance = 10000000
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_p_intron))) || (PyList_CheckExact(__pyx_v_p_intron))) {
        PyObject* sequence = __pyx_v_p_intron;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 386, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_21 = PyObject_GetIter(__pyx_v_p_intron); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_23 = Py_TYPE(__pyx_t_21)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_23(__pyx_t_21); if (unlikely(!__pyx_t_3)) goto __pyx_L24_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_1 = __pyx_t_23(__pyx_t_21); if (unlikely(!__pyx_t_1)) goto __pyx_L24_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_23(__pyx_t_21), 2) < 0) __PYX_ERR(0, 386, __pyx_L1_error)
        __pyx_t_23 = NULL;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        goto __pyx_L25_unpacking_done;
        __pyx_L24_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        __pyx_t_23 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 386, __pyx_L1_error)
        __pyx_L25_unpacking_done:;
      }
      __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_p_splice_start = __pyx_t_16;
      __pyx_v_p_splice_end = __pyx_t_4;

      /* "Mikado/scales/contrast.pyx":387
 *         for p_intron in prediction.introns:
 *             p_splice_start, p_splice_end = p_intron
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")             # <<<<<<<<<<<<<<
 *             curr_distance = 10000000
 *             best = -1
 */
      __pyx_t_17.__pyx_n = 5;
      __pyx_t_17.strict = 0;
      __pyx_t_17.contained_check = 0;
      __pyx_t_17.max_distance = __pyx_v_fuzzymatch;
      __pyx_t_17.n = 0x3E8;
      __pyx_t_17.value = __pyx_n_u_intron;
      __pyx_t_1 = ((struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_v_r_tree->__pyx_vtab)->find(__pyx_v_r_tree, (__pyx_v_p_splice_start - 1), (__pyx_v_p_splice_start + 1), 0, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_found, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":388
 *             p_splice_start, p_splice_end = p_intron
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")
 *             curr_distance = 10000000             # <<<<<<<<<<<<<<
 *             best = -1
 *             for r_intron in found:
 */
      __pyx_v_curr_distance = 0x989680;

      /* "Mikado/scales/contrast.pyx":389
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")
 *             curr_distance = 10000000
 *             best = -1             # <<<<<<<<<<<<<<
 *             for r_intron in found:
 *                 r_splice_start = r_intron.start
 */
      __pyx_v_best = -1;

      /* "Mikado/scales/contrast.pyx":390
 *             curr_distance = 10000000
 *             best = -1
 *             for r_intron in found:             # <<<<<<<<<<<<<<
 *                 r_splice_start = r_intron.start
 *                 distance = abs(r_splice_start - p_splice_start)
 */
      if (likely(PyList_CheckExact(__pyx_v_found)) || PyTuple_CheckExact(__pyx_v_found)) {
        __pyx_t_1 = __pyx_v_found; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_13 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_found); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 390, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 390, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 390, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_19(__pyx_t_1);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 390, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_r_intron, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Mikado/scales/contrast.pyx":391
 *             best = -1
 *             for r_intron in found:
 *                 r_splice_start = r_intron.start             # <<<<<<<<<<<<<<
 *                 distance = abs(r_splice_start - p_splice_start)
 *                 if distance <= min(fuzzymatch, curr_distance):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_r_intron, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 391, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_r_splice_start = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":392
 *             for r_intron in found:
 *                 r_splice_start = r_intron.start
 *                 distance = abs(r_splice_start - p_splice_start)             # <<<<<<<<<<<<<<
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_start
 */
        __pyx_t_4 = abs((__pyx_v_r_splice_start - __pyx_v_p_splice_start)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 392, __pyx_L1_error)
        __pyx_v_distance = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":393
 *                 r_splice_start = r_intron.start
 *                 distance = abs(r_splice_start - p_splice_start)
 *                 if distance <= min(fuzzymatch, curr_distance):             # <<<<<<<<<<<<<<
 *                     best = p_splice_start
 *                     curr_distance = distance
 */
        __pyx_t_10 = __pyx_v_curr_distance;
        __pyx_t_4 = __pyx_v_fuzzymatch;
        if (((__pyx_t_10 < __pyx_t_4) != 0)) {
          __pyx_t_9 = __pyx_t_10;
        } else {
          __pyx_t_9 = __pyx_t_4;
        }
        __pyx_t_11 = ((__pyx_v_distance <= __pyx_t_9) != 0);
        if (__pyx_t_11) {

          /* "Mikado/scales/contrast.pyx":394
 *                 distance = abs(r_splice_start - p_splice_start)
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_start             # <<<<<<<<<<<<<<
 *                     curr_distance = distance
 *             if best > 0:
 */
          __pyx_v_best = __pyx_v_p_splice_start;

          /* "Mikado/scales/contrast.pyx":395
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_start
 *                     curr_distance = distance             # <<<<<<<<<<<<<<
 *             if best > 0:
 *                 c_splices.add(best)
 */
          __pyx_v_curr_distance = __pyx_v_distance;

          /* "Mikado/scales/contrast.pyx":393
 *                 r_splice_start = r_intron.start
 *                 distance = abs(r_splice_start - p_splice_start)
 *                 if distance <= min(fuzzymatch, curr_distance):             # <<<<<<<<<<<<<<
 *                     best = p_splice_start
 *                     curr_distance = distance
 */
        }

        /* "Mikado/scales/contrast.pyx":390
 *             curr_distance = 10000000
 *             best = -1
 *             for r_intron in found:             # <<<<<<<<<<<<<<
 *                 r_splice_start = r_intron.start
 *                 distance = abs(r_splice_start - p_splice_start)
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":396
 *                     best = p_splice_start
 *                     curr_distance = distance
 *             if best > 0:             # <<<<<<<<<<<<<<
 *                 c_splices.add(best)
 *             best = -1
 */
      __pyx_t_11 = ((__pyx_v_best > 0) != 0);
      if (__pyx_t_11) {

        /* "Mikado/scales/contrast.pyx":397
 *                     curr_distance = distance
 *             if best > 0:
 *                 c_splices.add(best)             # <<<<<<<<<<<<<<
 *             best = -1
 *             curr_distance = 10000000
 */
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_15 = PySet_Add(__pyx_v_c_splices, __pyx_t_1); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 397, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Mikado/scales/contrast.pyx":396
 *                     best = p_splice_start
 *                     curr_distance = distance
 *             if best > 0:             # <<<<<<<<<<<<<<
 *                 c_splices.add(best)
 *             best = -1
 */
      }

      /* "Mikado/scales/contrast.pyx":398
 *             if best > 0:
 *                 c_splices.add(best)
 *             best = -1             # <<<<<<<<<<<<<<
 *             curr_distance = 10000000
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")
 */
      __pyx_v_best = -1;

      /* "Mikado/scales/contrast.pyx":399
 *                 c_splices.add(best)
 *             best = -1
 *             curr_distance = 10000000             # <<<<<<<<<<<<<<
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")
 *             for r_intron in found:
 */
      __pyx_v_curr_distance = 0x989680;

      /* "Mikado/scales/contrast.pyx":400
 *             best = -1
 *             curr_distance = 10000000
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")             # <<<<<<<<<<<<<<
 *             for r_intron in found:
 *                 r_splice_end = r_intron.end
 */
      __pyx_t_17.__pyx_n = 5;
      __pyx_t_17.strict = 0;
      __pyx_t_17.contained_check = 0;
      __pyx_t_17.max_distance = __pyx_v_fuzzymatch;
      __pyx_t_17.n = 0x3E8;
      __pyx_t_17.value = __pyx_n_u_intron;
      __pyx_t_1 = ((struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree *)__pyx_v_r_tree->__pyx_vtab)->find(__pyx_v_r_tree, (__pyx_v_p_splice_end - 1), __pyx_v_p_splice_end, 0, &__pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_found, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":401
 *             curr_distance = 10000000
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")
 *             for r_intron in found:             # <<<<<<<<<<<<<<
 *                 r_splice_end = r_intron.end
 *                 distance = abs(r_splice_end - p_splice_end)
 */
      if (likely(PyList_CheckExact(__pyx_v_found)) || PyTuple_CheckExact(__pyx_v_found)) {
        __pyx_t_1 = __pyx_v_found; __Pyx_INCREF(__pyx_t_1); __pyx_t_13 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_13 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_found); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 401, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_19(__pyx_t_1);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 401, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_r_intron, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Mikado/scales/contrast.pyx":402
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")
 *             for r_intron in found:
 *                 r_splice_end = r_intron.end             # <<<<<<<<<<<<<<
 *                 distance = abs(r_splice_end - p_splice_end)
 *                 if distance <= min(fuzzymatch, curr_distance):
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_r_intron, __pyx_n_s_end); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_r_splice_end = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":403
 *             for r_intron in found:
 *                 r_splice_end = r_intron.end
 *                 distance = abs(r_splice_end - p_splice_end)             # <<<<<<<<<<<<<<
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_end
 */
        __pyx_t_4 = abs((__pyx_v_r_splice_end - __pyx_v_p_splice_end)); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 403, __pyx_L1_error)
        __pyx_v_distance = __pyx_t_4;

        /* "Mikado/scales/contrast.pyx":404
 *                 r_splice_end = r_intron.end
 *                 distance = abs(r_splice_end - p_splice_end)
 *                 if distance <= min(fuzzymatch, curr_distance):             # <<<<<<<<<<<<<<
 *                     best = p_splice_end
 *                     curr_distance = distance
 */
        __pyx_t_9 = __pyx_v_curr_distance;
        __pyx_t_4 = __pyx_v_fuzzymatch;
        if (((__pyx_t_9 < __pyx_t_4) != 0)) {
          __pyx_t_10 = __pyx_t_9;
        } else {
          __pyx_t_10 = __pyx_t_4;
        }
        __pyx_t_11 = ((__pyx_v_distance <= __pyx_t_10) != 0);
        if (__pyx_t_11) {

          /* "Mikado/scales/contrast.pyx":405
 *                 distance = abs(r_splice_end - p_splice_end)
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_end             # <<<<<<<<<<<<<<
 *                     curr_distance = distance
 *             if best > 0:
 */
          __pyx_v_best = __pyx_v_p_splice_end;

          /* "Mikado/scales/contrast.pyx":406
 *                 if distance <= min(fuzzymatch, curr_distance):
 *                     best = p_splice_end
 *                     curr_distance = distance             # <<<<<<<<<<<<<<
 *             if best > 0:
 *                 c_splices.add(best)
 */
          __pyx_v_curr_distance = __pyx_v_distance;

          /* "Mikado/scales/contrast.pyx":404
 *                 r_splice_end = r_intron.end
 *                 distance = abs(r_splice_end - p_splice_end)
 *                 if distance <= min(fuzzymatch, curr_distance):             # <<<<<<<<<<<<<<
 *                     best = p_splice_end
 *                     curr_distance = distance
 */
        }

        /* "Mikado/scales/contrast.pyx":401
 *             curr_distance = 10000000
 *             found = r_tree.find(p_splice_end - 1, p_splice_end, 0, 0, fuzzymatch, 1000, "intron")
 *             for r_intron in found:             # <<<<<<<<<<<<<<
 *                 r_splice_end = r_intron.end
 *                 distance = abs(r_splice_end - p_splice_end)
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":407
 *                     best = p_splice_end
 *                     curr_distance = distance
 *             if best > 0:             # <<<<<<<<<<<<<<
 *                 c_splices.add(best)
 * 
 */
      __pyx_t_11 = ((__pyx_v_best > 0) != 0);
      if (__pyx_t_11) {

        /* "Mikado/scales/contrast.pyx":408
 *                     curr_distance = distance
 *             if best > 0:
 *                 c_splices.add(best)             # <<<<<<<<<<<<<<
 * 
 *         len_rsplices, len_psplices, junction_overlap = len(r_splices), len(p_splices), len(c_splices)
 */
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_15 = PySet_Add(__pyx_v_c_splices, __pyx_t_1); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 408, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Mikado/scales/contrast.pyx":407
 *                     best = p_splice_end
 *                     curr_distance = distance
 *             if best > 0:             # <<<<<<<<<<<<<<
 *                 c_splices.add(best)
 * 
 */
      }

      /* "Mikado/scales/contrast.pyx":385
 *     if p_exon_num > 1 and r_exon_num > 1:
 *         r_splices, p_splices = reference.splices, prediction.splices
 *         for p_intron in prediction.introns:             # <<<<<<<<<<<<<<
 *             p_splice_start, p_splice_end = p_intron
 *             found = r_tree.find(p_splice_start - 1, p_splice_start + 1, 0, 0, fuzzymatch, 1000, "intron")
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Mikado/scales/contrast.pyx":410
 *                 c_splices.add(best)
 * 
 *         len_rsplices, len_psplices, junction_overlap = len(r_splices), len(p_splices), len(c_splices)             # <<<<<<<<<<<<<<
 *         junction_recall = junction_overlap / len_rsplices
 *         junction_precision = junction_overlap / len_psplices
 */
    if (unlikely(__pyx_v_r_splices == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 410, __pyx_L1_error)
    }
    __pyx_t_18 = PySet_GET_SIZE(__pyx_v_r_splices); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
    if (unlikely(__pyx_v_p_splices == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 410, __pyx_L1_error)
    }
    __pyx_t_13 = PySet_GET_SIZE(__pyx_v_p_splices); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
    __pyx_t_24 = PySet_GET_SIZE(__pyx_v_c_splices); if (unlikely(__pyx_t_24 == ((Py_ssize_t)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
    __pyx_v_len_rsplices = __pyx_t_18;
    __pyx_v_len_psplices = __pyx_t_13;
    __pyx_v_junction_overlap = __pyx_t_24;

    /* "Mikado/scales/contrast.pyx":411
 * 
 *         len_rsplices, len_psplices, junction_overlap = len(r_splices), len(p_splices), len(c_splices)
 *         junction_recall = junction_overlap / len_rsplices             # <<<<<<<<<<<<<<
 *         junction_precision = junction_overlap / len_psplices
 *         junction_f1 = calc_f1(junction_recall, junction_precision)
 */
    __pyx_v_junction_recall = (((double)__pyx_v_junction_overlap) / __pyx_v_len_rsplices);

    /* "Mikado/scales/contrast.pyx":412
 *         len_rsplices, len_psplices, junction_overlap = len(r_splices), len(p_splices), len(c_splices)
 *         junction_recall = junction_overlap / len_rsplices
 *         junction_precision = junction_overlap / len_psplices             # <<<<<<<<<<<<<<
 *         junction_f1 = calc_f1(junction_recall, junction_precision)
 * 
 */
    __pyx_v_junction_precision = (((double)__pyx_v_junction_overlap) / __pyx_v_len_psplices);

    /* "Mikado/scales/contrast.pyx":413
 *         junction_recall = junction_overlap / len_rsplices
 *         junction_precision = junction_overlap / len_psplices
 *         junction_f1 = calc_f1(junction_recall, junction_precision)             # <<<<<<<<<<<<<<
 * 
 *     elif p_exon_num == r_exon_num == 1 and nucl_f1 > 0:
 */
    __pyx_v_junction_f1 = __pyx_f_6Mikado_9utilities_2f1_calc_f1(__pyx_v_junction_recall, __pyx_v_junction_precision, 0);

    /* "Mikado/scales/contrast.pyx":383
 *     junction_overlap = 0
 *     c_splices = set()
 *     if p_exon_num > 1 and r_exon_num > 1:             # <<<<<<<<<<<<<<
 *         r_splices, p_splices = reference.splices, prediction.splices
 *         for p_intron in prediction.introns:
 */
    goto __pyx_L19;
  }

  /* "Mikado/scales/contrast.pyx":415
 *         junction_f1 = calc_f1(junction_recall, junction_precision)
 * 
 *     elif p_exon_num == r_exon_num == 1 and nucl_f1 > 0:             # <<<<<<<<<<<<<<
 *         # junction_overlap = junction_f1 = junction_precision = junction_recall = 1
 *         junction_f1 = junction_precision = junction_recall = 1
 */
  __pyx_t_22 = (__pyx_v_p_exon_num == __pyx_v_r_exon_num);
  if (__pyx_t_22) {
    __pyx_t_22 = (__pyx_v_r_exon_num == 1);
  }
  __pyx_t_12 = (__pyx_t_22 != 0);
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L34_bool_binop_done;
  }
  __pyx_t_12 = ((__pyx_v_nucl_f1 > 0.0) != 0);
  __pyx_t_11 = __pyx_t_12;
  __pyx_L34_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":417
 *     elif p_exon_num == r_exon_num == 1 and nucl_f1 > 0:
 *         # junction_overlap = junction_f1 = junction_precision = junction_recall = 1
 *         junction_f1 = junction_precision = junction_recall = 1             # <<<<<<<<<<<<<<
 *     else:
 *         # junction_overlap = junction_f1 = junction_precision = junction_recall = 0
 */
    __pyx_v_junction_f1 = 1.0;
    __pyx_v_junction_precision = 1.0;
    __pyx_v_junction_recall = 1.0;

    /* "Mikado/scales/contrast.pyx":415
 *         junction_f1 = calc_f1(junction_recall, junction_precision)
 * 
 *     elif p_exon_num == r_exon_num == 1 and nucl_f1 > 0:             # <<<<<<<<<<<<<<
 *         # junction_overlap = junction_f1 = junction_precision = junction_recall = 1
 *         junction_f1 = junction_precision = junction_recall = 1
 */
    goto __pyx_L19;
  }

  /* "Mikado/scales/contrast.pyx":420
 *     else:
 *         # junction_overlap = junction_f1 = junction_precision = junction_recall = 0
 *         junction_f1 = junction_precision = junction_recall = 0             # <<<<<<<<<<<<<<
 * 
 *     ccode = ""
 */
  /*else*/ {
    __pyx_v_junction_f1 = 0.0;
    __pyx_v_junction_precision = 0.0;
    __pyx_v_junction_recall = 0.0;
  }
  __pyx_L19:;

  /* "Mikado/scales/contrast.pyx":422
 *         junction_f1 = junction_precision = junction_recall = 0
 * 
 *     ccode = ""             # <<<<<<<<<<<<<<
 *     distance = 0
 *     if junction_f1 == 1 and p_exon_num > 1:
 */
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_v_ccode = __pyx_kp_u_;

  /* "Mikado/scales/contrast.pyx":423
 * 
 *     ccode = ""
 *     distance = 0             # <<<<<<<<<<<<<<
 *     if junction_f1 == 1 and p_exon_num > 1:
 *         if p_strand == r_strand:
 */
  __pyx_v_distance = 0;

  /* "Mikado/scales/contrast.pyx":424
 *     ccode = ""
 *     distance = 0
 *     if junction_f1 == 1 and p_exon_num > 1:             # <<<<<<<<<<<<<<
 *         if p_strand == r_strand:
 *             ccode = "="  # We have recovered all the junctions
 */
  __pyx_t_12 = ((__pyx_v_junction_f1 == 1.0) != 0);
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L37_bool_binop_done;
  }
  __pyx_t_12 = ((__pyx_v_p_exon_num > 1) != 0);
  __pyx_t_11 = __pyx_t_12;
  __pyx_L37_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":425
 *     distance = 0
 *     if junction_f1 == 1 and p_exon_num > 1:
 *         if p_strand == r_strand:             # <<<<<<<<<<<<<<
 *             ccode = "="  # We have recovered all the junctions
 *         else:
 */
    __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_p_strand, __pyx_v_r_strand, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 425, __pyx_L1_error)
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {

      /* "Mikado/scales/contrast.pyx":426
 *     if junction_f1 == 1 and p_exon_num > 1:
 *         if p_strand == r_strand:
 *             ccode = "="  # We have recovered all the junctions             # <<<<<<<<<<<<<<
 *         else:
 *             ccode = "c"  # We will set this to x at the end of the function
 */
      __Pyx_INCREF(__pyx_kp_u__3);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_kp_u__3);

      /* "Mikado/scales/contrast.pyx":425
 *     distance = 0
 *     if junction_f1 == 1 and p_exon_num > 1:
 *         if p_strand == r_strand:             # <<<<<<<<<<<<<<
 *             ccode = "="  # We have recovered all the junctions
 *         else:
 */
      goto __pyx_L39;
    }

    /* "Mikado/scales/contrast.pyx":428
 *             ccode = "="  # We have recovered all the junctions
 *         else:
 *             ccode = "c"  # We will set this to x at the end of the function             # <<<<<<<<<<<<<<
 * 
 *     elif junction_f1 == 1 and nucl_f1 >= 0.80:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_u_c);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_c);
    }
    __pyx_L39:;

    /* "Mikado/scales/contrast.pyx":424
 *     ccode = ""
 *     distance = 0
 *     if junction_f1 == 1 and p_exon_num > 1:             # <<<<<<<<<<<<<<
 *         if p_strand == r_strand:
 *             ccode = "="  # We have recovered all the junctions
 */
    goto __pyx_L36;
  }

  /* "Mikado/scales/contrast.pyx":430
 *             ccode = "c"  # We will set this to x at the end of the function
 * 
 *     elif junction_f1 == 1 and nucl_f1 >= 0.80:             # <<<<<<<<<<<<<<
 *         reference_exon = reference.exons[0]
 *         ccode = "_"  # We have recovered all the junctions
 */
  __pyx_t_11 = ((__pyx_v_junction_f1 == 1.0) != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_12 = __pyx_t_11;
    goto __pyx_L40_bool_binop_done;
  }
  __pyx_t_11 = ((__pyx_v_nucl_f1 >= 0.80) != 0);
  __pyx_t_12 = __pyx_t_11;
  __pyx_L40_bool_binop_done:;
  if (__pyx_t_12) {

    /* "Mikado/scales/contrast.pyx":431
 * 
 *     elif junction_f1 == 1 and nucl_f1 >= 0.80:
 *         reference_exon = reference.exons[0]             # <<<<<<<<<<<<<<
 *         ccode = "_"  # We have recovered all the junctions
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_reference_exon, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Mikado/scales/contrast.pyx":432
 *     elif junction_f1 == 1 and nucl_f1 >= 0.80:
 *         reference_exon = reference.exons[0]
 *         ccode = "_"  # We have recovered all the junctions             # <<<<<<<<<<<<<<
 * 
 *     # Outside the transcript - polymerase run-on
 */
    __Pyx_INCREF(__pyx_n_u__2);
    __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u__2);

    /* "Mikado/scales/contrast.pyx":430
 *             ccode = "c"  # We will set this to x at the end of the function
 * 
 *     elif junction_f1 == 1 and nucl_f1 >= 0.80:             # <<<<<<<<<<<<<<
 *         reference_exon = reference.exons[0]
 *         ccode = "_"  # We have recovered all the junctions
 */
    goto __pyx_L36;
  }

  /* "Mikado/scales/contrast.pyx":435
 * 
 *     # Outside the transcript - polymerase run-on
 *     elif p_start > r_end or p_end < r_start:             # <<<<<<<<<<<<<<
 *         if r_strand == p_strand:
 *             ccode = "p"
 */
  __pyx_t_11 = ((__pyx_v_p_start > __pyx_v_r_end) != 0);
  if (!__pyx_t_11) {
  } else {
    __pyx_t_12 = __pyx_t_11;
    goto __pyx_L42_bool_binop_done;
  }
  __pyx_t_11 = ((__pyx_v_p_end < __pyx_v_r_start) != 0);
  __pyx_t_12 = __pyx_t_11;
  __pyx_L42_bool_binop_done:;
  if (__pyx_t_12) {

    /* "Mikado/scales/contrast.pyx":436
 *     # Outside the transcript - polymerase run-on
 *     elif p_start > r_end or p_end < r_start:
 *         if r_strand == p_strand:             # <<<<<<<<<<<<<<
 *             ccode = "p"
 *         else:
 */
    __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_v_r_strand, __pyx_v_p_strand, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_12 != 0);
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":437
 *     elif p_start > r_end or p_end < r_start:
 *         if r_strand == p_strand:
 *             ccode = "p"             # <<<<<<<<<<<<<<
 *         else:
 *             ccode = "P"
 */
      __Pyx_INCREF(__pyx_n_u_p);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_p);

      /* "Mikado/scales/contrast.pyx":436
 *     # Outside the transcript - polymerase run-on
 *     elif p_start > r_end or p_end < r_start:
 *         if r_strand == p_strand:             # <<<<<<<<<<<<<<
 *             ccode = "p"
 *         else:
 */
      goto __pyx_L44;
    }

    /* "Mikado/scales/contrast.pyx":439
 *             ccode = "p"
 *         else:
 *             ccode = "P"             # <<<<<<<<<<<<<<
 *         distance = long_max(p_start - r_end, r_start - p_end)
 * 
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_u_P);
      __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_P);
    }
    __pyx_L44:;

    /* "Mikado/scales/contrast.pyx":440
 *         else:
 *             ccode = "P"
 *         distance = long_max(p_start - r_end, r_start - p_end)             # <<<<<<<<<<<<<<
 * 
 *     if ccode == "":
 */
    __pyx_v_distance = __pyx_f_6Mikado_6scales_8contrast_long_max((__pyx_v_p_start - __pyx_v_r_end), (__pyx_v_r_start - __pyx_v_p_end));

    /* "Mikado/scales/contrast.pyx":435
 * 
 *     # Outside the transcript - polymerase run-on
 *     elif p_start > r_end or p_end < r_start:             # <<<<<<<<<<<<<<
 *         if r_strand == p_strand:
 *             ccode = "p"
 */
  }
  __pyx_L36:;

  /* "Mikado/scales/contrast.pyx":442
 *         distance = long_max(p_start - r_end, r_start - p_end)
 * 
 *     if ccode == "":             # <<<<<<<<<<<<<<
 *         # stats = (nucl_recall, nucl_precision, nucl_f1,
 *         #         exon_recall, exon_precision, exon_f1,
 */
  __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_ccode, __pyx_kp_u_, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 442, __pyx_L1_error)
  __pyx_t_12 = (__pyx_t_11 != 0);
  if (__pyx_t_12) {

    /* "Mikado/scales/contrast.pyx":446
 *         #         exon_recall, exon_precision, exon_f1,
 *         #          junction_recall, junction_precision, junction_f1)
 *         stats[0] = nucl_recall             # <<<<<<<<<<<<<<
 *         stats[1] = nucl_precision
 *         stats[2] = nucl_f1
 */
    (__pyx_v_stats[0]) = __pyx_v_nucl_recall;

    /* "Mikado/scales/contrast.pyx":447
 *         #          junction_recall, junction_precision, junction_f1)
 *         stats[0] = nucl_recall
 *         stats[1] = nucl_precision             # <<<<<<<<<<<<<<
 *         stats[2] = nucl_f1
 *         stats[3] = exon_recall
 */
    (__pyx_v_stats[1]) = __pyx_v_nucl_precision;

    /* "Mikado/scales/contrast.pyx":448
 *         stats[0] = nucl_recall
 *         stats[1] = nucl_precision
 *         stats[2] = nucl_f1             # <<<<<<<<<<<<<<
 *         stats[3] = exon_recall
 *         stats[4] = exon_precision
 */
    (__pyx_v_stats[2]) = __pyx_v_nucl_f1;

    /* "Mikado/scales/contrast.pyx":449
 *         stats[1] = nucl_precision
 *         stats[2] = nucl_f1
 *         stats[3] = exon_recall             # <<<<<<<<<<<<<<
 *         stats[4] = exon_precision
 *         stats[5] = exon_f1
 */
    (__pyx_v_stats[3]) = __pyx_v_exon_recall;

    /* "Mikado/scales/contrast.pyx":450
 *         stats[2] = nucl_f1
 *         stats[3] = exon_recall
 *         stats[4] = exon_precision             # <<<<<<<<<<<<<<
 *         stats[5] = exon_f1
 *         stats[6] = junction_recall
 */
    (__pyx_v_stats[4]) = __pyx_v_exon_precision;

    /* "Mikado/scales/contrast.pyx":451
 *         stats[3] = exon_recall
 *         stats[4] = exon_precision
 *         stats[5] = exon_f1             # <<<<<<<<<<<<<<
 *         stats[6] = junction_recall
 *         stats[7] = junction_precision
 */
    (__pyx_v_stats[5]) = __pyx_v_exon_f1;

    /* "Mikado/scales/contrast.pyx":452
 *         stats[4] = exon_precision
 *         stats[5] = exon_f1
 *         stats[6] = junction_recall             # <<<<<<<<<<<<<<
 *         stats[7] = junction_precision
 *         stats[8] = junction_f1
 */
    (__pyx_v_stats[6]) = __pyx_v_junction_recall;

    /* "Mikado/scales/contrast.pyx":453
 *         stats[5] = exon_f1
 *         stats[6] = junction_recall
 *         stats[7] = junction_precision             # <<<<<<<<<<<<<<
 *         stats[8] = junction_f1
 * 
 */
    (__pyx_v_stats[7]) = __pyx_v_junction_precision;

    /* "Mikado/scales/contrast.pyx":454
 *         stats[6] = junction_recall
 *         stats[7] = junction_precision
 *         stats[8] = junction_f1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_stats[8]) = __pyx_v_junction_f1;

    /* "Mikado/scales/contrast.pyx":457
 * 
 * 
 *         if double_min(p_exon_num, r_exon_num) > 1:             # <<<<<<<<<<<<<<
 *             ccode = __assign_multiexonic_ccode(prediction, reference,
 *                                                nucl_overlap, stats, fuzzymatch)
 */
    __pyx_t_12 = ((__pyx_f_6Mikado_6scales_8contrast_double_min(__pyx_v_p_exon_num, __pyx_v_r_exon_num) > 1.0) != 0);
    if (__pyx_t_12) {

      /* "Mikado/scales/contrast.pyx":458
 * 
 *         if double_min(p_exon_num, r_exon_num) > 1:
 *             ccode = __assign_multiexonic_ccode(prediction, reference,             # <<<<<<<<<<<<<<
 *                                                nucl_overlap, stats, fuzzymatch)
 *         else:
 */
      __pyx_t_25.__pyx_n = 1;
      __pyx_t_25.fuzzymatch = __pyx_v_fuzzymatch;
      __pyx_t_1 = __pyx_f_6Mikado_6scales_8contrast___assign_multiexonic_ccode(__pyx_v_prediction, __pyx_v_reference, __pyx_v_nucl_overlap, __pyx_v_stats, &__pyx_t_25); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_ccode, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "Mikado/scales/contrast.pyx":457
 * 
 * 
 *         if double_min(p_exon_num, r_exon_num) > 1:             # <<<<<<<<<<<<<<
 *             ccode = __assign_multiexonic_ccode(prediction, reference,
 *                                                nucl_overlap, stats, fuzzymatch)
 */
      goto __pyx_L46;
    }

    /* "Mikado/scales/contrast.pyx":461
 *                                                nucl_overlap, stats, fuzzymatch)
 *         else:
 *             ccode = __assign_monoexonic_ccode(prediction, reference,             # <<<<<<<<<<<<<<
 *                                               nucl_overlap, stats)
 * 
 */
    /*else*/ {

      /* "Mikado/scales/contrast.pyx":462
 *         else:
 *             ccode = __assign_monoexonic_ccode(prediction, reference,
 *                                               nucl_overlap, stats)             # <<<<<<<<<<<<<<
 * 
 *     if (p_strand != r_strand):
 */
      __pyx_t_1 = __pyx_f_6Mikado_6scales_8contrast___assign_monoexonic_ccode(__pyx_v_prediction, __pyx_v_reference, __pyx_v_nucl_overlap, __pyx_v_stats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_ccode, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;
    }
    __pyx_L46:;

    /* "Mikado/scales/contrast.pyx":442
 *         distance = long_max(p_start - r_end, r_start - p_end)
 * 
 *     if ccode == "":             # <<<<<<<<<<<<<<
 *         # stats = (nucl_recall, nucl_precision, nucl_f1,
 *         #         exon_recall, exon_precision, exon_f1,
 */
  }

  /* "Mikado/scales/contrast.pyx":464
 *                                               nucl_overlap, stats)
 * 
 *     if (p_strand != r_strand):             # <<<<<<<<<<<<<<
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 */
  __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_v_p_strand, __pyx_v_r_strand, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
  __pyx_t_11 = (__pyx_t_12 != 0);
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":465
 * 
 *     if (p_strand != r_strand):
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):             # <<<<<<<<<<<<<<
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 *                 ccode = "x"  # "x{0}".format(ccode)
 */
    __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_v_p_strand, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
    __pyx_t_22 = (__pyx_t_12 != 0);
    if (!__pyx_t_22) {
      goto __pyx_L50_next_or;
    } else {
    }
    __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_v_r_strand, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
    __pyx_t_12 = (__pyx_t_22 != 0);
    if (!__pyx_t_12) {
    } else {
      __pyx_t_11 = __pyx_t_12;
      goto __pyx_L49_bool_binop_done;
    }
    __pyx_L50_next_or:;
    __pyx_t_12 = ((__pyx_v_strict_strandedness == 1) != 0);
    __pyx_t_11 = __pyx_t_12;
    __pyx_L49_bool_binop_done:;
    if (__pyx_t_11) {

      /* "Mikado/scales/contrast.pyx":466
 *     if (p_strand != r_strand):
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):
 *             if ccode in ("e", "g", "c", "m", "_", "C"):             # <<<<<<<<<<<<<<
 *                 ccode = "x"  # "x{0}".format(ccode)
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):
 */
      __Pyx_INCREF(__pyx_v_ccode);
      __pyx_t_26 = __pyx_v_ccode;
      __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_e, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_12 != 0);
      if (!__pyx_t_22) {
      } else {
        __pyx_t_11 = __pyx_t_22;
        goto __pyx_L53_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_g, Py_EQ)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_22 != 0);
      if (!__pyx_t_12) {
      } else {
        __pyx_t_11 = __pyx_t_12;
        goto __pyx_L53_bool_binop_done;
      }
      __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_12 != 0);
      if (!__pyx_t_22) {
      } else {
        __pyx_t_11 = __pyx_t_22;
        goto __pyx_L53_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_m, Py_EQ)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_22 != 0);
      if (!__pyx_t_12) {
      } else {
        __pyx_t_11 = __pyx_t_12;
        goto __pyx_L53_bool_binop_done;
      }
      __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u__2, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_12 != 0);
      if (!__pyx_t_22) {
      } else {
        __pyx_t_11 = __pyx_t_22;
        goto __pyx_L53_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_C, Py_EQ)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 466, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_22 != 0);
      __pyx_t_11 = __pyx_t_12;
      __pyx_L53_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Mikado/scales/contrast.pyx":467
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 *                 ccode = "x"  # "x{0}".format(ccode)             # <<<<<<<<<<<<<<
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):
 *                 ccode = "X"  # "X{0}".format(ccode)
 */
        __Pyx_INCREF(__pyx_n_u_x);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_x);

        /* "Mikado/scales/contrast.pyx":466
 *     if (p_strand != r_strand):
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):
 *             if ccode in ("e", "g", "c", "m", "_", "C"):             # <<<<<<<<<<<<<<
 *                 ccode = "x"  # "x{0}".format(ccode)
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):
 */
        goto __pyx_L52;
      }

      /* "Mikado/scales/contrast.pyx":468
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 *                 ccode = "x"  # "x{0}".format(ccode)
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):             # <<<<<<<<<<<<<<
 *                 ccode = "X"  # "X{0}".format(ccode)
 *         reference_exon = None
 */
      __Pyx_INCREF(__pyx_v_ccode);
      __pyx_t_26 = __pyx_v_ccode;
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_u, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_11 != 0);
      if (__pyx_t_22) {
      } else {
        __pyx_t_12 = __pyx_t_22;
        goto __pyx_L59_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_i, Py_NE)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_22 != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L59_bool_binop_done;
      }
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_I, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_11 != 0);
      if (__pyx_t_22) {
      } else {
        __pyx_t_12 = __pyx_t_22;
        goto __pyx_L59_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_p, Py_NE)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_22 != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_12 = __pyx_t_11;
        goto __pyx_L59_bool_binop_done;
      }
      __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_P, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_22 = (__pyx_t_11 != 0);
      if (__pyx_t_22) {
      } else {
        __pyx_t_12 = __pyx_t_22;
        goto __pyx_L59_bool_binop_done;
      }
      __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_t_26, __pyx_n_u_x, Py_NE)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 468, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_22 != 0);
      __pyx_t_12 = __pyx_t_11;
      __pyx_L59_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
      __pyx_t_11 = (__pyx_t_12 != 0);
      if (__pyx_t_11) {

        /* "Mikado/scales/contrast.pyx":469
 *                 ccode = "x"  # "x{0}".format(ccode)
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):
 *                 ccode = "X"  # "X{0}".format(ccode)             # <<<<<<<<<<<<<<
 *         reference_exon = None
 * 
 */
        __Pyx_INCREF(__pyx_n_u_X);
        __Pyx_DECREF_SET(__pyx_v_ccode, __pyx_n_u_X);

        /* "Mikado/scales/contrast.pyx":468
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 *                 ccode = "x"  # "x{0}".format(ccode)
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):             # <<<<<<<<<<<<<<
 *                 ccode = "X"  # "X{0}".format(ccode)
 *         reference_exon = None
 */
      }
      __pyx_L52:;

      /* "Mikado/scales/contrast.pyx":465
 * 
 *     if (p_strand != r_strand):
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):             # <<<<<<<<<<<<<<
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 *                 ccode = "x"  # "x{0}".format(ccode)
 */
    }

    /* "Mikado/scales/contrast.pyx":470
 *             elif ccode not in ("u", "i", "I", "p", "P", "x"):
 *                 ccode = "X"  # "X{0}".format(ccode)
 *         reference_exon = None             # <<<<<<<<<<<<<<
 * 
 *     location = "{}:{}..{}".format(reference.chrom,
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_reference_exon, Py_None);

    /* "Mikado/scales/contrast.pyx":464
 *                                               nucl_overlap, stats)
 * 
 *     if (p_strand != r_strand):             # <<<<<<<<<<<<<<
 *         if (p_strand != "" and r_strand != "") or (strict_strandedness == 1):
 *             if ccode in ("e", "g", "c", "m", "_", "C"):
 */
  }

  /* "Mikado/scales/contrast.pyx":472
 *         reference_exon = None
 * 
 *     location = "{}:{}..{}".format(reference.chrom,             # <<<<<<<<<<<<<<
 *                                   min(reference.start, prediction.start),
 *                                   max(reference.end, prediction.end))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__4, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_chrom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "Mikado/scales/contrast.pyx":473
 * 
 *     location = "{}:{}..{}".format(reference.chrom,
 *                                   min(reference.start, prediction.start),             # <<<<<<<<<<<<<<
 *                                   max(reference.end, prediction.end))
 * 
 */
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_start); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_start); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_28 = PyObject_RichCompare(__pyx_t_21, __pyx_t_20, Py_LT); __Pyx_XGOTREF(__pyx_t_28); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_28); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  if (__pyx_t_11) {
    __Pyx_INCREF(__pyx_t_21);
    __pyx_t_27 = __pyx_t_21;
  } else {
    __Pyx_INCREF(__pyx_t_20);
    __pyx_t_27 = __pyx_t_20;
  }
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "Mikado/scales/contrast.pyx":474
 *     location = "{}:{}..{}".format(reference.chrom,
 *                                   min(reference.start, prediction.start),
 *                                   max(reference.end, prediction.end))             # <<<<<<<<<<<<<<
 * 
 *     if reference.parent == [] or reference.parent is None:
 */
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_end); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_end); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_29 = PyObject_RichCompare(__pyx_t_21, __pyx_t_20, Py_GT); __Pyx_XGOTREF(__pyx_t_29); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 474, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_29); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  if (__pyx_t_11) {
    __Pyx_INCREF(__pyx_t_21);
    __pyx_t_28 = __pyx_t_21;
  } else {
    __Pyx_INCREF(__pyx_t_20);
    __pyx_t_28 = __pyx_t_20;
  }
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_21)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_21);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_21, __pyx_t_3, __pyx_t_27, __pyx_t_28};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_21, __pyx_t_3, __pyx_t_27, __pyx_t_28};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  } else
  #endif
  {
    __pyx_t_20 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__pyx_t_21) {
      __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_21); __pyx_t_21 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_4, __pyx_t_3);
    __Pyx_INCREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_4, __pyx_t_27);
    __Pyx_INCREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_4, __pyx_t_28);
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_location = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":476
 *                                   max(reference.end, prediction.end))
 * 
 *     if reference.parent == [] or reference.parent is None:             # <<<<<<<<<<<<<<
 *         ref_parent = "NA"
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_20); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_20); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (!__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L66_bool_binop_done;
  }
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_parent); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_12 = (__pyx_t_20 == Py_None);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_22 = (__pyx_t_12 != 0);
  __pyx_t_11 = __pyx_t_22;
  __pyx_L66_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":477
 * 
 *     if reference.parent == [] or reference.parent is None:
 *         ref_parent = "NA"             # <<<<<<<<<<<<<<
 *     else:
 *         ref_parent = ",".join([str(_) for _ in reference.parent])
 */
    __Pyx_INCREF(__pyx_n_u_NA);
    __pyx_v_ref_parent = __pyx_n_u_NA;

    /* "Mikado/scales/contrast.pyx":476
 *                                   max(reference.end, prediction.end))
 * 
 *     if reference.parent == [] or reference.parent is None:             # <<<<<<<<<<<<<<
 *         ref_parent = "NA"
 *     else:
 */
    goto __pyx_L65;
  }

  /* "Mikado/scales/contrast.pyx":479
 *         ref_parent = "NA"
 *     else:
 *         ref_parent = ",".join([str(_) for _ in reference.parent])             # <<<<<<<<<<<<<<
 * 
 *     if prediction.parent == [] or prediction.parent is None:
 */
  /*else*/ {
    { /* enter inner scope */
      __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 479, __pyx_L70_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_parent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L70_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_24 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_24 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L70_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 479, __pyx_L70_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_24); __Pyx_INCREF(__pyx_t_2); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 479, __pyx_L70_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L70_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_24); __Pyx_INCREF(__pyx_t_2); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 479, __pyx_L70_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L70_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_14(__pyx_t_1);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 479, __pyx_L70_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v__, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_7genexpr__pyx_v__); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L70_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_20, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 479, __pyx_L70_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v__); __pyx_7genexpr__pyx_v__ = 0;
      goto __pyx_L73_exit_scope;
      __pyx_L70_error:;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v__); __pyx_7genexpr__pyx_v__ = 0;
      goto __pyx_L1_error;
      __pyx_L73_exit_scope:;
    } /* exit inner scope */
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__5, __pyx_t_20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_v_ref_parent = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L65:;

  /* "Mikado/scales/contrast.pyx":481
 *         ref_parent = ",".join([str(_) for _ in reference.parent])
 * 
 *     if prediction.parent == [] or prediction.parent is None:             # <<<<<<<<<<<<<<
 *         pred_parent = "NA"
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_20, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_22 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_22) {
  } else {
    __pyx_t_11 = __pyx_t_22;
    goto __pyx_L75_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_parent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_22 = (__pyx_t_2 == Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = (__pyx_t_22 != 0);
  __pyx_t_11 = __pyx_t_12;
  __pyx_L75_bool_binop_done:;
  if (__pyx_t_11) {

    /* "Mikado/scales/contrast.pyx":482
 * 
 *     if prediction.parent == [] or prediction.parent is None:
 *         pred_parent = "NA"             # <<<<<<<<<<<<<<
 *     else:
 *         pred_parent = ",".join([str(_) for _ in prediction.parent])
 */
    __Pyx_INCREF(__pyx_n_u_NA);
    __pyx_v_pred_parent = __pyx_n_u_NA;

    /* "Mikado/scales/contrast.pyx":481
 *         ref_parent = ",".join([str(_) for _ in reference.parent])
 * 
 *     if prediction.parent == [] or prediction.parent is None:             # <<<<<<<<<<<<<<
 *         pred_parent = "NA"
 *     else:
 */
    goto __pyx_L74;
  }

  /* "Mikado/scales/contrast.pyx":484
 *         pred_parent = "NA"
 *     else:
 *         pred_parent = ",".join([str(_) for _ in prediction.parent])             # <<<<<<<<<<<<<<
 * 
 *     result = ResultStorer(reference.id,
 */
  /*else*/ {
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L79_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_parent); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 484, __pyx_L79_error)
      __Pyx_GOTREF(__pyx_t_20);
      if (likely(PyList_CheckExact(__pyx_t_20)) || PyTuple_CheckExact(__pyx_t_20)) {
        __pyx_t_1 = __pyx_t_20; __Pyx_INCREF(__pyx_t_1); __pyx_t_24 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_24 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L79_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 484, __pyx_L79_error)
      }
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_20 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_24); __Pyx_INCREF(__pyx_t_20); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L79_error)
            #else
            __pyx_t_20 = PySequence_ITEM(__pyx_t_1, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 484, __pyx_L79_error)
            __Pyx_GOTREF(__pyx_t_20);
            #endif
          } else {
            if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_24); __Pyx_INCREF(__pyx_t_20); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L79_error)
            #else
            __pyx_t_20 = PySequence_ITEM(__pyx_t_1, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 484, __pyx_L79_error)
            __Pyx_GOTREF(__pyx_t_20);
            #endif
          }
        } else {
          __pyx_t_20 = __pyx_t_14(__pyx_t_1);
          if (unlikely(!__pyx_t_20)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 484, __pyx_L79_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_20);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v__, __pyx_t_20);
        __pyx_t_20 = 0;
        __pyx_t_20 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_8genexpr1__pyx_v__); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 484, __pyx_L79_error)
        __Pyx_GOTREF(__pyx_t_20);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_20))) __PYX_ERR(0, 484, __pyx_L79_error)
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v__); __pyx_8genexpr1__pyx_v__ = 0;
      goto __pyx_L82_exit_scope;
      __pyx_L79_error:;
      __Pyx_XDECREF(__pyx_8genexpr1__pyx_v__); __pyx_8genexpr1__pyx_v__ = 0;
      goto __pyx_L1_error;
      __pyx_L82_exit_scope:;
    } /* exit inner scope */
    __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_pred_parent = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L74:;

  /* "Mikado/scales/contrast.pyx":486
 *         pred_parent = ",".join([str(_) for _ in prediction.parent])
 * 
 *     result = ResultStorer(reference.id,             # <<<<<<<<<<<<<<
 *                           ref_parent,
 *                           ccode, prediction.id,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ResultStorer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_id); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);

  /* "Mikado/scales/contrast.pyx":488
 *     result = ResultStorer(reference.id,
 *                           ref_parent,
 *                           ccode, prediction.id,             # <<<<<<<<<<<<<<
 *                           pred_parent,
 *                           len(prediction.exons),
 */
  __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_id); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);

  /* "Mikado/scales/contrast.pyx":490
 *                           ccode, prediction.id,
 *                           pred_parent,
 *                           len(prediction.exons),             # <<<<<<<<<<<<<<
 *                           len(reference.exons),
 *                           # Nucleotide stats
 */
  __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_v_prediction, __pyx_n_s_exons); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_24 = PyObject_Length(__pyx_t_27); if (unlikely(__pyx_t_24 == ((Py_ssize_t)-1))) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = PyInt_FromSsize_t(__pyx_t_24); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);

  /* "Mikado/scales/contrast.pyx":491
 *                           pred_parent,
 *                           len(prediction.exons),
 *                           len(reference.exons),             # <<<<<<<<<<<<<<
 *                           # Nucleotide stats
 *                           round(nucl_precision * 100, 2),
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_reference, __pyx_n_s_exons); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_24 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_24 == ((Py_ssize_t)-1))) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_24); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "Mikado/scales/contrast.pyx":493
 *                           len(reference.exons),
 *                           # Nucleotide stats
 *                           round(nucl_precision * 100, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * nucl_recall, 2),
 *                           round(100 * nucl_f1, 2),
 */
  __pyx_t_21 = PyFloat_FromDouble((__pyx_v_nucl_precision * 100.0)); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_t_21);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_int_2);
  __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_29, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "Mikado/scales/contrast.pyx":494
 *                           # Nucleotide stats
 *                           round(nucl_precision * 100, 2),
 *                           round(100 * nucl_recall, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * nucl_f1, 2),
 *                           # Junction stats
 */
  __pyx_t_29 = PyFloat_FromDouble((100.0 * __pyx_v_nucl_recall)); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_t_29);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_int_2);
  __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_30, NULL); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "Mikado/scales/contrast.pyx":495
 *                           round(nucl_precision * 100, 2),
 *                           round(100 * nucl_recall, 2),
 *                           round(100 * nucl_f1, 2),             # <<<<<<<<<<<<<<
 *                           # Junction stats
 *                           round(junction_precision * 100, 2),
 */
  __pyx_t_30 = PyFloat_FromDouble((100.0 * __pyx_v_nucl_f1)); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_30);
  PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_t_30);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_int_2);
  __pyx_t_30 = 0;
  __pyx_t_30 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_31, NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "Mikado/scales/contrast.pyx":497
 *                           round(100 * nucl_f1, 2),
 *                           # Junction stats
 *                           round(junction_precision * 100, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * junction_recall, 2),
 *                           round(100 * junction_f1, 2),
 */
  __pyx_t_31 = PyFloat_FromDouble((__pyx_v_junction_precision * 100.0)); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_31);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_t_31);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_int_2);
  __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_32, NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "Mikado/scales/contrast.pyx":498
 *                           # Junction stats
 *                           round(junction_precision * 100, 2),
 *                           round(100 * junction_recall, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * junction_f1, 2),
 *                           # Exonic stats
 */
  __pyx_t_32 = PyFloat_FromDouble((100.0 * __pyx_v_junction_recall)); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_32);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_t_32);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_int_2);
  __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_33, NULL); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "Mikado/scales/contrast.pyx":499
 *                           round(junction_precision * 100, 2),
 *                           round(100 * junction_recall, 2),
 *                           round(100 * junction_f1, 2),             # <<<<<<<<<<<<<<
 *                           # Exonic stats
 *                           round(exon_precision * 100, 2),
 */
  __pyx_t_33 = PyFloat_FromDouble((100.0 * __pyx_v_junction_f1)); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_t_33);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_int_2);
  __pyx_t_33 = 0;
  __pyx_t_33 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_34, NULL); if (unlikely(!__pyx_t_33)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "Mikado/scales/contrast.pyx":501
 *                           round(100 * junction_f1, 2),
 *                           # Exonic stats
 *                           round(exon_precision * 100, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * exon_recall, 2),
 *                           round(100 * exon_f1, 2),
 */
  __pyx_t_34 = PyFloat_FromDouble((__pyx_v_exon_precision * 100.0)); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_t_34);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_int_2);
  __pyx_t_34 = 0;
  __pyx_t_34 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_35, NULL); if (unlikely(!__pyx_t_34)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "Mikado/scales/contrast.pyx":502
 *                           # Exonic stats
 *                           round(exon_precision * 100, 2),
 *                           round(100 * exon_recall, 2),             # <<<<<<<<<<<<<<
 *                           round(100 * exon_f1, 2),
 *                           distance,
 */
  __pyx_t_35 = PyFloat_FromDouble((100.0 * __pyx_v_exon_recall)); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_t_35);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_int_2);
  __pyx_t_35 = 0;
  __pyx_t_35 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_36, NULL); if (unlikely(!__pyx_t_35)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "Mikado/scales/contrast.pyx":503
 *                           round(exon_precision * 100, 2),
 *                           round(100 * exon_recall, 2),
 *                           round(100 * exon_f1, 2),             # <<<<<<<<<<<<<<
 *                           distance,
 *                           location)
 */
  __pyx_t_36 = PyFloat_FromDouble((100.0 * __pyx_v_exon_f1)); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_36);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_int_2);
  __pyx_t_36 = 0;
  __pyx_t_36 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_37, NULL); if (unlikely(!__pyx_t_36)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "Mikado/scales/contrast.pyx":504
 *                           round(100 * exon_recall, 2),
 *                           round(100 * exon_f1, 2),
 *                           distance,             # <<<<<<<<<<<<<<
 *                           location)
 * 
 */
  __pyx_t_37 = __Pyx_PyInt_From_long(__pyx_v_distance); if (unlikely(!__pyx_t_37)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);

  /* "Mikado/scales/contrast.pyx":505
 *                           round(100 * exon_f1, 2),
 *                           distance,
 *                           location)             # <<<<<<<<<<<<<<
 * 
 *     if ccode == "":
 */
  __pyx_t_38 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_38 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_38)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_38);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[19] = {__pyx_t_38, __pyx_t_20, __pyx_v_ref_parent, __pyx_v_ccode, __pyx_t_28, __pyx_v_pred_parent, __pyx_t_27, __pyx_t_3, __pyx_t_21, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_v_location};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 18+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_38); __pyx_t_38 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
    __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
    __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[19] = {__pyx_t_38, __pyx_t_20, __pyx_v_ref_parent, __pyx_v_ccode, __pyx_t_28, __pyx_v_pred_parent, __pyx_t_27, __pyx_t_3, __pyx_t_21, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32, __pyx_t_33, __pyx_t_34, __pyx_t_35, __pyx_t_36, __pyx_t_37, __pyx_v_location};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 18+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_38); __pyx_t_38 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
    __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
    __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
    __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  } else
  #endif
  {
    __pyx_t_39 = PyTuple_New(18+__pyx_t_4); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_39);
    if (__pyx_t_38) {
      __Pyx_GIVEREF(__pyx_t_38); PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_t_38); __pyx_t_38 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_39, 0+__pyx_t_4, __pyx_t_20);
    __Pyx_INCREF(__pyx_v_ref_parent);
    __Pyx_GIVEREF(__pyx_v_ref_parent);
    PyTuple_SET_ITEM(__pyx_t_39, 1+__pyx_t_4, __pyx_v_ref_parent);
    __Pyx_INCREF(__pyx_v_ccode);
    __Pyx_GIVEREF(__pyx_v_ccode);
    PyTuple_SET_ITEM(__pyx_t_39, 2+__pyx_t_4, __pyx_v_ccode);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_39, 3+__pyx_t_4, __pyx_t_28);
    __Pyx_INCREF(__pyx_v_pred_parent);
    __Pyx_GIVEREF(__pyx_v_pred_parent);
    PyTuple_SET_ITEM(__pyx_t_39, 4+__pyx_t_4, __pyx_v_pred_parent);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_39, 5+__pyx_t_4, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_39, 6+__pyx_t_4, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_21);
    PyTuple_SET_ITEM(__pyx_t_39, 7+__pyx_t_4, __pyx_t_21);
    __Pyx_GIVEREF(__pyx_t_29);
    PyTuple_SET_ITEM(__pyx_t_39, 8+__pyx_t_4, __pyx_t_29);
    __Pyx_GIVEREF(__pyx_t_30);
    PyTuple_SET_ITEM(__pyx_t_39, 9+__pyx_t_4, __pyx_t_30);
    __Pyx_GIVEREF(__pyx_t_31);
    PyTuple_SET_ITEM(__pyx_t_39, 10+__pyx_t_4, __pyx_t_31);
    __Pyx_GIVEREF(__pyx_t_32);
    PyTuple_SET_ITEM(__pyx_t_39, 11+__pyx_t_4, __pyx_t_32);
    __Pyx_GIVEREF(__pyx_t_33);
    PyTuple_SET_ITEM(__pyx_t_39, 12+__pyx_t_4, __pyx_t_33);
    __Pyx_GIVEREF(__pyx_t_34);
    PyTuple_SET_ITEM(__pyx_t_39, 13+__pyx_t_4, __pyx_t_34);
    __Pyx_GIVEREF(__pyx_t_35);
    PyTuple_SET_ITEM(__pyx_t_39, 14+__pyx_t_4, __pyx_t_35);
    __Pyx_GIVEREF(__pyx_t_36);
    PyTuple_SET_ITEM(__pyx_t_39, 15+__pyx_t_4, __pyx_t_36);
    __Pyx_GIVEREF(__pyx_t_37);
    PyTuple_SET_ITEM(__pyx_t_39, 16+__pyx_t_4, __pyx_t_37);
    __Pyx_INCREF(__pyx_v_location);
    __Pyx_GIVEREF(__pyx_v_location);
    PyTuple_SET_ITEM(__pyx_t_39, 17+__pyx_t_4, __pyx_v_location);
    __pyx_t_20 = 0;
    __pyx_t_28 = 0;
    __pyx_t_27 = 0;
    __pyx_t_3 = 0;
    __pyx_t_21 = 0;
    __pyx_t_29 = 0;
    __pyx_t_30 = 0;
    __pyx_t_31 = 0;
    __pyx_t_32 = 0;
    __pyx_t_33 = 0;
    __pyx_t_34 = 0;
    __pyx_t_35 = 0;
    __pyx_t_36 = 0;
    __pyx_t_37 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":507
 *                           location)
 * 
 *     if ccode == "":             # <<<<<<<<<<<<<<
 *         raise ValueError("Ccode is null;\n{0}".format(repr(result)))
 * 
 */
  __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_v_ccode, __pyx_kp_u_, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 507, __pyx_L1_error)
  __pyx_t_12 = (__pyx_t_11 != 0);
  if (unlikely(__pyx_t_12)) {

    /* "Mikado/scales/contrast.pyx":508
 * 
 *     if ccode == "":
 *         raise ValueError("Ccode is null;\n{0}".format(repr(result)))             # <<<<<<<<<<<<<<
 * 
 *     return (result, reference_exon)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Ccode_is_null_0, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_39 = PyObject_Repr(__pyx_v_result); if (unlikely(!__pyx_t_39)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_39);
    __pyx_t_37 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_37 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_37)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_37);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_37) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_37, __pyx_t_39) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_39);
    __Pyx_XDECREF(__pyx_t_37); __pyx_t_37 = 0;
    __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 508, __pyx_L1_error)

    /* "Mikado/scales/contrast.pyx":507
 *                           location)
 * 
 *     if ccode == "":             # <<<<<<<<<<<<<<
 *         raise ValueError("Ccode is null;\n{0}".format(repr(result)))
 * 
 */
  }

  /* "Mikado/scales/contrast.pyx":510
 *         raise ValueError("Ccode is null;\n{0}".format(repr(result)))
 * 
 *     return (result, reference_exon)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_result);
  __Pyx_GIVEREF(__pyx_v_result);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_result);
  __Pyx_INCREF(__pyx_v_reference_exon);
  __Pyx_GIVEREF(__pyx_v_reference_exon);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_reference_exon);
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Mikado/scales/contrast.pyx":239
 * @cython.profile(True)
 * @cython.cdivision(True)
 * cpdef tuple compare(prediction, reference, bint lenient=False, bint strict_strandedness=False, int fuzzymatch=0):             # <<<<<<<<<<<<<<
 * 
 *     """Cython function to compare two transcripts and determine a ccode.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_AddTraceback("Mikado.scales.contrast.compare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pred_exons);
  __Pyx_XDECREF(__pyx_v___ref_exons);
  __Pyx_XDECREF((PyObject *)__pyx_v_r_tree);
  __Pyx_XDECREF(__pyx_v_r_splices);
  __Pyx_XDECREF(__pyx_v_p_splices);
  __Pyx_XDECREF(__pyx_v_c_splices);
  __Pyx_XDECREF(__pyx_v_r_strand);
  __Pyx_XDECREF(__pyx_v_p_strand);
  __Pyx_XDECREF(__pyx_v_exon);
  __Pyx_XDECREF(__pyx_v_other_exon);
  __Pyx_XDECREF(__pyx_v_reference_exon);
  __Pyx_XDECREF(__pyx_v_p_intron);
  __Pyx_XDECREF(__pyx_v_found);
  __Pyx_XDECREF(__pyx_v_r_intron);
  __Pyx_XDECREF(__pyx_v_ccode);
  __Pyx_XDECREF(__pyx_v_location);
  __Pyx_XDECREF(__pyx_v_ref_parent);
  __Pyx_XDECREF(__pyx_v_pred_parent);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v__);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6Mikado_6scales_8contrast_1compare(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6Mikado_6scales_8contrast_compare[] = "Cython function to compare two transcripts and determine a ccode.\n\n    :param prediction: the transcript query\n    :type prediction: Transcript\n\n    :param reference: the reference transcript against which we desire to calculate the ccode and other stats.\n    :type reference: Transcript\n    \n    :param lenient: a boolean flag that indicates whether the exon-level features should be calculated leniently or not.\n    :type lenient: bool\n\n    :rtype (ResultStorer, (int,int)) | (ResultStorer, None)\n\n    Available ccodes (from Cufflinks documentation):\n\n    - =    Complete intron chain match\n    - c    Contained (perfect junction recall and precision, imperfect recall)\n    - j    Potentially novel isoform (fragment): at least one splice junction is shared with a reference transcript\n    - e    Single exon transfrag overlapping a reference exon and at least 10 bp of a reference intron, indicating a possible pre-mRNA fragment.\n    - i    A *monoexonic* transfrag falling entirely within a reference intron\n    - o    Generic exonic overlap with a reference transcript\n    - p    Possible polymerase run-on fragment (within 2Kbases of a reference transcript)\n    - u    Unknown, intergenic transcript\n    - x    Exonic overlap with reference on the opposite strand (class codes e, o, m, c, _)\n    - X    Overlap on the opposite strand, with some junctions in common (probably a serious mistake, unless non-canonical splicing junctions are involved).\n\n    Please note that the description for i is changed from Cufflinks.\n\n    We also provide the following additional classifications:\n\n    - f    gene fusion - in this case, this ccode will be followed by the ccodes of the matches for each gene, separated by comma\n    - _    Complete match, for monoexonic transcripts (nucleotide F1>=80% - i.e. min(precision,recall)>=66.7%\n    - m    Exon overlap between two monoexonic transcripts\n    - n    Potential extension of the reference - we have added new splice jun""ctions *outside* the boundaries of the transcript itself\n    - C    Contained transcript with overextensions on either side (perfect junction recall, imperfect nucleotide specificity)\n    - J    Potentially novel isoform, where all the known junctions have been confirmed and we have added others as well *externally*\n    - I    *multiexonic* transcript falling completely inside a known transcript\n    - h    AS event in which at least a couple of introns overlaps but without any\n           junction in common.\n    - G    Reverse generic overlap - the reference is monoexonic while the prediction isn't\n    - P    Possible polymerase run-on fragment\n    - g   Monoexonic overlap - the prediction is monoexonic and the reference is multiexonic\n    (within 2K bases of a reference transcript), on the opposite strand\n\n    This is a class method, and can therefore be used outside of a class instance.\n    ";
static PyObject *__pyx_pw_6Mikado_6scales_8contrast_1compare(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_prediction = 0;
  PyObject *__pyx_v_reference = 0;
  int __pyx_v_lenient;
  int __pyx_v_strict_strandedness;
  int __pyx_v_fuzzymatch;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compare (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_prediction,&__pyx_n_s_reference,&__pyx_n_s_lenient,&__pyx_n_s_strict_strandedness,&__pyx_n_s_fuzzymatch,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_prediction)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("compare", 0, 2, 5, 1); __PYX_ERR(0, 239, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lenient);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strict_strandedness);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fuzzymatch);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compare") < 0)) __PYX_ERR(0, 239, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_prediction = values[0];
    __pyx_v_reference = values[1];
    if (values[2]) {
      __pyx_v_lenient = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_lenient == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
    } else {
      __pyx_v_lenient = ((int)0);
    }
    if (values[3]) {
      __pyx_v_strict_strandedness = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_strict_strandedness == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
    } else {
      __pyx_v_strict_strandedness = ((int)0);
    }
    if (values[4]) {
      __pyx_v_fuzzymatch = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_fuzzymatch == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
    } else {
      __pyx_v_fuzzymatch = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compare", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 239, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Mikado.scales.contrast.compare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6Mikado_6scales_8contrast_compare(__pyx_self, __pyx_v_prediction, __pyx_v_reference, __pyx_v_lenient, __pyx_v_strict_strandedness, __pyx_v_fuzzymatch);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6Mikado_6scales_8contrast_compare(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_prediction, PyObject *__pyx_v_reference, int __pyx_v_lenient, int __pyx_v_strict_strandedness, int __pyx_v_fuzzymatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarations
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6Mikado_6scales_8contrast_compare __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("compare", 0);
  __Pyx_TraceCall("compare (wrapper)", __pyx_f[0], 239, 0, __PYX_ERR(0, 239, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.lenient = __pyx_v_lenient;
  __pyx_t_2.strict_strandedness = __pyx_v_strict_strandedness;
  __pyx_t_2.fuzzymatch = __pyx_v_fuzzymatch;
  __pyx_t_1 = __pyx_f_6Mikado_6scales_8contrast_compare(__pyx_v_prediction, __pyx_v_reference, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Mikado.scales.contrast.compare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_TraceReturn(__pyx_r, 0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"compare", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6Mikado_6scales_8contrast_1compare, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6Mikado_6scales_8contrast_compare},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_contrast(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_contrast},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "contrast",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_n_u_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 1, 0, 1},
  {&__pyx_kp_u_Ccode_is_null_0, __pyx_k_Ccode_is_null_0, sizeof(__pyx_k_Ccode_is_null_0), 0, 1, 0, 0},
  {&__pyx_n_u_G, __pyx_k_G, sizeof(__pyx_k_G), 0, 1, 0, 1},
  {&__pyx_n_u_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 1, 0, 1},
  {&__pyx_n_u_J, __pyx_k_J, sizeof(__pyx_k_J), 0, 1, 0, 1},
  {&__pyx_kp_u_Luca_Venturini, __pyx_k_Luca_Venturini, sizeof(__pyx_k_Luca_Venturini), 0, 1, 0, 0},
  {&__pyx_n_u_NA, __pyx_k_NA, sizeof(__pyx_k_NA), 0, 1, 0, 1},
  {&__pyx_n_u_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 1, 0, 1},
  {&__pyx_n_s_ResultStorer, __pyx_k_ResultStorer, sizeof(__pyx_k_ResultStorer), 0, 0, 1, 1},
  {&__pyx_n_s_Transcript, __pyx_k_Transcript, sizeof(__pyx_k_Transcript), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_u_X, __pyx_k_X, sizeof(__pyx_k_X), 0, 1, 0, 1},
  {&__pyx_n_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 1},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_n_s_author, __pyx_k_author, sizeof(__pyx_k_author), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_cdna_length, __pyx_k_cdna_length, sizeof(__pyx_k_cdna_length), 0, 0, 1, 1},
  {&__pyx_n_s_chrom, __pyx_k_chrom, sizeof(__pyx_k_chrom), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_difference, __pyx_k_difference, sizeof(__pyx_k_difference), 0, 0, 1, 1},
  {&__pyx_n_u_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 1, 0, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_u_exon, __pyx_k_exon, sizeof(__pyx_k_exon), 0, 1, 0, 1},
  {&__pyx_n_s_exon_num, __pyx_k_exon_num, sizeof(__pyx_k_exon_num), 0, 0, 1, 1},
  {&__pyx_n_s_exons, __pyx_k_exons, sizeof(__pyx_k_exons), 0, 0, 1, 1},
  {&__pyx_n_s_finalize, __pyx_k_finalize, sizeof(__pyx_k_finalize), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fuzzymatch, __pyx_k_fuzzymatch, sizeof(__pyx_k_fuzzymatch), 0, 0, 1, 1},
  {&__pyx_n_u_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 1, 0, 1},
  {&__pyx_n_u_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 1, 0, 1},
  {&__pyx_n_u_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 1, 0, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {&__pyx_n_u_intron, __pyx_k_intron, sizeof(__pyx_k_intron), 0, 1, 0, 1},
  {&__pyx_n_s_introns, __pyx_k_introns, sizeof(__pyx_k_introns), 0, 0, 1, 1},
  {&__pyx_n_u_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 1, 0, 1},
  {&__pyx_n_s_lenient, __pyx_k_lenient, sizeof(__pyx_k_lenient), 0, 0, 1, 1},
  {&__pyx_n_u_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 1, 0, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_u_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 1, 0, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_u_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 1, 0, 1},
  {&__pyx_n_u_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 1, 0, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_prediction, __pyx_k_prediction, sizeof(__pyx_k_prediction), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_u_rI, __pyx_k_rI, sizeof(__pyx_k_rI), 0, 1, 0, 1},
  {&__pyx_n_s_reference, __pyx_k_reference, sizeof(__pyx_k_reference), 0, 0, 1, 1},
  {&__pyx_n_s_resultstorer, __pyx_k_resultstorer, sizeof(__pyx_k_resultstorer), 0, 0, 1, 1},
  {&__pyx_n_u_ri, __pyx_k_ri, sizeof(__pyx_k_ri), 0, 1, 0, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_s_segmenttree, __pyx_k_segmenttree, sizeof(__pyx_k_segmenttree), 0, 0, 1, 1},
  {&__pyx_n_s_splices, __pyx_k_splices, sizeof(__pyx_k_splices), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_strand, __pyx_k_strand, sizeof(__pyx_k_strand), 0, 0, 1, 1},
  {&__pyx_n_s_strict_strandedness, __pyx_k_strict_strandedness, sizeof(__pyx_k_strict_strandedness), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_transcripts, __pyx_k_transcripts, sizeof(__pyx_k_transcripts), 0, 0, 1, 1},
  {&__pyx_n_u_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 1, 0, 1},
  {&__pyx_n_u_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 176, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 493, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 508, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PySet_Type_copy.type = (PyObject*)&PySet_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("compare", (void (*)(void))__pyx_f_6Mikado_6scales_8contrast_compare, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6Mikado_6scales_8contrast_compare *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__assign_multiexonic_ccode", (void (*)(void))__pyx_f_6Mikado_6scales_8contrast___assign_multiexonic_ccode, "PyObject *(PyObject *, PyObject *, long, double *, struct __pyx_opt_args_6Mikado_6scales_8contrast___assign_multiexonic_ccode *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__assign_monoexonic_ccode", (void (*)(void))__pyx_f_6Mikado_6scales_8contrast___assign_monoexonic_ccode, "PyObject *(PyObject *, PyObject *, long, double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("Mikado.utilities.intervaltree"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6Mikado_9utilities_12intervaltree_Interval = __Pyx_ImportType(__pyx_t_1, "Mikado.utilities.intervaltree", "Interval", sizeof(struct __pyx_obj_6Mikado_9utilities_12intervaltree_Interval), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6Mikado_9utilities_12intervaltree_Interval) __PYX_ERR(1, 3, __pyx_L1_error)
  __pyx_vtabptr_6Mikado_9utilities_12intervaltree_Interval = (struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_Interval*)__Pyx_GetVtable(__pyx_ptype_6Mikado_9utilities_12intervaltree_Interval->tp_dict); if (unlikely(!__pyx_vtabptr_6Mikado_9utilities_12intervaltree_Interval)) __PYX_ERR(1, 3, __pyx_L1_error)
  __pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalNode = __Pyx_ImportType(__pyx_t_1, "Mikado.utilities.intervaltree", "IntervalNode", sizeof(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalNode) __PYX_ERR(1, 11, __pyx_L1_error)
  __pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalNode = (struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalNode*)__Pyx_GetVtable(__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalNode->tp_dict); if (unlikely(!__pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalNode)) __PYX_ERR(1, 11, __pyx_L1_error)
  __pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree = __Pyx_ImportType(__pyx_t_1, "Mikado.utilities.intervaltree", "IntervalTree", sizeof(struct __pyx_obj_6Mikado_9utilities_12intervaltree_IntervalTree), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree) __PYX_ERR(1, 35, __pyx_L1_error)
  __pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalTree = (struct __pyx_vtabstruct_6Mikado_9utilities_12intervaltree_IntervalTree*)__Pyx_GetVtable(__pyx_ptype_6Mikado_9utilities_12intervaltree_IntervalTree->tp_dict); if (unlikely(!__pyx_vtabptr_6Mikado_9utilities_12intervaltree_IntervalTree)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("Mikado.utilities.overlap"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "c_overlap", (void (**)(void))&__pyx_f_6Mikado_9utilities_7overlap_c_overlap, "long (long, long, long, long, long, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Mikado.utilities.f1"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "calc_f1", (void (**)(void))&__pyx_f_6Mikado_9utilities_2f1_calc_f1, "double (double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcontrast(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcontrast(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_contrast(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_contrast(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_contrast(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'contrast' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_contrast(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("contrast", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_Mikado__scales__contrast) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "Mikado.scales.contrast")) {
      if (unlikely(PyDict_SetItemString(modules, "Mikado.scales.contrast", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "Mikado/scales/contrast.pyx":1
 * from .resultstorer import ResultStorer             # <<<<<<<<<<<<<<
 * from ..transcripts import Transcript
 * from ..utilities.overlap cimport c_overlap
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ResultStorer);
  __Pyx_GIVEREF(__pyx_n_s_ResultStorer);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ResultStorer);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_resultstorer, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ResultStorer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ResultStorer, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Mikado/scales/contrast.pyx":2
 * from .resultstorer import ResultStorer
 * from ..transcripts import Transcript             # <<<<<<<<<<<<<<
 * from ..utilities.overlap cimport c_overlap
 * from ..utilities.intervaltree cimport Interval, IntervalTree
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Transcript);
  __Pyx_GIVEREF(__pyx_n_s_Transcript);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Transcript);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_transcripts, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Transcript); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Transcript, __pyx_t_2) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Mikado/scales/contrast.pyx":9
 * 
 * 
 * __author__ = 'Luca Venturini'             # <<<<<<<<<<<<<<
 * 
 * cdef inline long long_max(long a, long b): return a if a >= b else b
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_author, __pyx_kp_u_Luca_Venturini) < 0) __PYX_ERR(0, 9, __pyx_L1_error)

  /* "Mikado/scales/contrast.pyx":1
 * from .resultstorer import ResultStorer             # <<<<<<<<<<<<<<
 * from ..transcripts import Transcript
 * from ..utilities.overlap cimport c_overlap
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init Mikado.scales.contrast", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init Mikado.scales.contrast");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* Profile */
#if CYTHON_PROFILE
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno) {
    PyObject *type, *value, *traceback;
    int retval;
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        if (*code == NULL) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_d,                  /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
#if PY_VERSION_HEX < 0x030400B1
    } else {
        (*frame)->f_tstate = tstate;
#endif
    }
    __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
    retval = 1;
    tstate->tracing++;
    __Pyx_SetTracing(tstate, 0);
    __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
    #if CYTHON_TRACE
    if (tstate->c_tracefunc)
        retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
    if (retval && tstate->c_profilefunc)
    #endif
        retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
    __Pyx_SetTracing(tstate, (tstate->c_profilefunc || (CYTHON_TRACE && tstate->c_tracefunc)));
    tstate->tracing--;
    if (retval) {
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
        return __Pyx_IsTracing(tstate, 0, 0) && retval;
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
        return -1;
    }
}
static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyCodeObject *py_code = 0;
#if PY_MAJOR_VERSION >= 3
    py_code = PyCode_NewEmpty(srcfile, funcname, firstlineno);
    if (likely(py_code)) {
        py_code->co_flags |= CO_OPTIMIZED | CO_NEWLOCALS;
    }
#else
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    py_funcname = PyString_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    py_srcfile = PyString_FromString(srcfile);
    if (unlikely(!py_srcfile)) goto bad;
    py_code = PyCode_New(
        0,
        0,
        0,
        CO_OPTIMIZED | CO_NEWLOCALS,
        __pyx_empty_bytes,     /*PyObject *code,*/
        __pyx_empty_tuple,     /*PyObject *consts,*/
        __pyx_empty_tuple,     /*PyObject *names,*/
        __pyx_empty_tuple,     /*PyObject *varnames,*/
        __pyx_empty_tuple,     /*PyObject *freevars,*/
        __pyx_empty_tuple,     /*PyObject *cellvars,*/
        py_srcfile,       /*PyObject *filename,*/
        py_funcname,      /*PyObject *name,*/
        firstlineno,
        __pyx_empty_bytes      /*PyObject *lnotab*/
    );
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
#endif
    return py_code;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    (void)is_set;
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        (void)orig_length;
        (void)ppos;
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_remove */
static int __Pyx_PySet_RemoveNotFound(PyObject *set, PyObject *key, int found) {
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    if (likely(found == 0)) {
        PyObject *tup;
        tup = PyTuple_Pack(1, key);
        if (!tup)
            return -1;
        PyErr_SetObject(PyExc_KeyError, tup);
        Py_DECREF(tup);
        return -1;
    }
    return found;
}
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found != 1)) {
        return __Pyx_PySet_RemoveNotFound(set, key, found);
    }
    return 0;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE short __Pyx_PyInt_As_short(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const short neg_one = (short) -1, const_zero = (short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(short) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(short, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (short) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case  1: __PYX_VERIFY_RETURN_INT(short, digit, digits[0])
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 2 * PyLong_SHIFT) {
                            return (short) (((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 3 * PyLong_SHIFT) {
                            return (short) (((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) >= 4 * PyLong_SHIFT) {
                            return (short) (((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (short) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(short) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (short) 0;
                case -1: __PYX_VERIFY_RETURN_INT(short, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(short,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(short) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(short) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                            return (short) ((((((short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(short) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(short) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                            return (short) ((((((((short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(short) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) (((short)-1)*(((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(short) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(short) - 1 > 4 * PyLong_SHIFT) {
                            return (short) ((((((((((short)digits[3]) << PyLong_SHIFT) | (short)digits[2]) << PyLong_SHIFT) | (short)digits[1]) << PyLong_SHIFT) | (short)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(short) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(short) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(short, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            short val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (short) -1;
        }
    } else {
        short val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (short) -1;
        val = __Pyx_PyInt_As_short(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to short");
    return (short) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to short");
    return (short) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
